/*
 * dma_engine.c
 *
 *  Created on: Aug 25, 2020
 *      Author: tom
 */
#include <asm/byteorder.h>
#include <asm/cacheflush.h>

#include "flipcie_regs.h"
#include "dma_engine.h"
#include "flipcie_ioctl.h"

static int32_t create_hw_descriptors(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap, DMA_DESCRIPTOR_LIST *pList, int32_t iNumDescriptors);
static int32_t alloc_desc_pages(fli_cam_data_t *pDev, uint32_t uiNumPages, DMA_DESC_INFO *pDescInfo);
static void free_desc_pages(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap);
static void DumpDescriptors(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap);
static void clear_poll_buffers(volatile XDMA_POLL_WB *pPollBuff);

int dma_engine_probe(struct pci_dev *pPciDev,fli_cam_data_t *pDev)
{

	int       ret = 0;

	dev_dbg(&(pPciDev->dev), "%s: Entered \n", __FUNCTION__);

	pDev->XDmaRegs.pRegs = (XDMA_ENGINE_REGS*)(pDev->bar1.hdl + OFFSET_C2H);   // BAR 1
	pDev->XDmaRegs.pSgdma = (XDMA_SGDMA_REGS*)(pDev->bar1.hdl + OFFSET_SGDMA); // BAR 1

	// Stop any existing DMA operations
	iowrite32(XDMA_CR_RUN_BIT, (void *)&(pDev->XDmaRegs.pRegs->controlW1C));
	iowrite32(XDMA_PC_AUTO | XDMA_PC_CLEAR, (void *)&(pDev->XDmaRegs.pRegs->perfCtrl));


	//pDev->pXDmaWBVirtAddr = (XDMA_POLL_WB *)pci_alloc_consistent(pPciDev, sizeof(XDMA_POLL_WB), &(pDev->uiPollPhysAddr));
	pDev->pXDmaWBVirtAddr = (XDMA_POLL_WB *)dma_alloc_coherent(&pPciDev->dev, sizeof(XDMA_POLL_WB), &(pDev->uiPollPhysAddr), GFP_KERNEL);
	if (pDev->pXDmaWBVirtAddr)
	{
		iowrite32(cpu_to_le32(PCI_DMA_LO(pDev->uiPollPhysAddr)), (void *)&(pDev->XDmaRegs.pRegs->pollModeWbLo));
		iowrite32(cpu_to_le32(PCI_DMA_HI(pDev->uiPollPhysAddr)), (void *)&(pDev->XDmaRegs.pRegs->pollModeWbHi));

		clear_poll_buffers(pDev->pXDmaWBVirtAddr);
	}
	else
	{
		pr_err("%s:%s: Error allocating poll write-back buffer\n", DEVICE_NAME, __FUNCTION__);
		ret = -ENOMEM;
	}

	return ret;
}

/// Clean of dma related items when a device is "removed", opposite of probe
int dma_engine_remove(struct pci_dev *pPciDev,fli_cam_data_t *pDev)
{

	if ((NULL == pPciDev) || (NULL == pDev))
		return(0);

   dev_dbg(&(pPciDev->dev), " %s : Entered\n", __FUNCTION__);

   if (pDev->pXDmaWBVirtAddr)
   {
	   //pci_free_consistent (pPciDev, sizeof(XDMA_POLL_WB), (void *)pDev->pXDmaWBVirtAddr, pDev->uiPollPhysAddr);
	   dma_free_coherent (&pPciDev->dev, sizeof(XDMA_POLL_WB), (void *)pDev->pXDmaWBVirtAddr, pDev->uiPollPhysAddr);
      pDev->pXDmaWBVirtAddr = NULL;
   }

   return 0;
}

int dma_engine_start_xfer(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	if ((NULL == pDev) || (NULL == pSGMap))
		return(-1);

	//dev_dbg(&(pDev->pci_devp->dev), "%s: Dumping dma regs before start ...\n", __FUNCTION__);
	//dma_engine_dump_regs(pDev);
	// write the address of the first SG descriptor
	// to the SG registers
	iowrite32(cpu_to_le32(PCI_DMA_LO(pSGMap->descInfo.uiDescPhysp)), pDev->bar1.hdl + GET_REG_ADDR(REG_SGD_DESC_LOW));
	iowrite32(cpu_to_le32(PCI_DMA_HI(pSGMap->descInfo.uiDescPhysp)), pDev->bar1.hdl + GET_REG_ADDR(REG_SGD_DESC_HIGH));
	iowrite32(XDMA_DESC_CONTROL_GET_ADJ_DESC_NUM(pSGMap->pHWDescriptors->control), pDev->bar1.hdl + GET_REG_ADDR(REG_SGD_DESC_ADJACENT));

	// The initialization of the Poll Write back information
	// was done in the probe function above.  All we need to do is
	// clear the old write back info for this new transaction
	clear_poll_buffers(pDev->pXDmaWBVirtAddr);

	// Kick off the DMA-
	// Clear and start performance counters on RUN, stop when final descriptor is completed.
	iowrite32(XDMA_PC_AUTO | XDMA_PC_RUN, (void *)&(pDev->XDmaRegs.pRegs->perfCtrl));
	iowrite32(XDMA_CR_NO_METADATA | XDMA_CR_POLL_MODE | XDMA_CR_RUN_BIT | XDMA_CR_IE_ALL, (void *)&(pDev->XDmaRegs.pRegs->controlW1S));

	//dev_dbg(&(pDev->pci_devp->dev), "%s: Dumping dma regs after start...\n", __FUNCTION__);
	//dma_engine_dump_regs(pDev);

	return(0);

}

int dma_engine_stop_xfer(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	if (NULL == pDev)
		return(-1);

	if (pSGMap)
	{
		dev_dbg(&(pDev->pci_devp->dev), "%s: syncing dma %d...\n", __FUNCTION__,pSGMap->sgTable.nents);
		//pci_dma_sync_sg_for_cpu(pDev->pci_devp, pSGMap->sgTable.sgl, pSGMap->sgTable.nents, DMA_FROM_DEVICE);
		dma_sync_sg_for_cpu(&pDev->pci_devp->dev, pSGMap->sgTable.sgl, pSGMap->sgTable.nents, DMA_FROM_DEVICE);
	}

	iowrite32(XDMA_CR_RUN_BIT, (void *)&(pDev->XDmaRegs.pRegs->controlW1C));

	return(0);

}

int dma_engine_is_xfer_complete(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	int32_t iResult;
	unsigned long uiCompleted;

	// Not complete
	iResult= 0;

	uiCompleted = pDev->pXDmaWBVirtAddr->uiCompletedDescCount;
    if (uiCompleted & XDMA_WB_ERR_MASK)
    {
		pr_err("%s: DMA Error 0x%lx, count= 0x%lx (0x%lx)\n", __FUNCTION__,(uiCompleted & 0x7F000000) >> 24, uiCompleted & XDMA_WB_COUNT_MASK, uiCompleted);
    	iResult= -1;
    }
    if (iResult >= 0)
	{
		// Are we done???
		uiCompleted= uiCompleted & XDMA_WB_COUNT_MASK;
		if (uiCompleted >= pSGMap->uiNumContigDescriptors)
			iResult= uiCompleted;
    }

    return(iResult);
}

void dma_engine_dump_regs(fli_cam_data_t *pDev)
{
	unsigned long uiCompleted;

	dev_dbg(&(pDev->pci_devp->dev), "%s: Write back uiCompletedDescCount 0x%x\n", __FUNCTION__,pDev->pXDmaWBVirtAddr->uiCompletedDescCount);
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->completedDescCount 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->completedDescCount)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->identifier 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->identifier)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->control 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->control)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->statusRC 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->statusRC)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->pollModeWbLo 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->pollModeWbLo)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->pollModeWbHi 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->pollModeWbHi)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->perfCtrl 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->perfCtrl)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->perfCycLo 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->perfCycLo)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->perfCycHi 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->perfCycHi)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->perfDatLo 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->perfDatLo)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pRegs->perfDatHi 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pRegs->perfDatHi)));

	dev_dbg(&(pDev->pci_devp->dev), "%s: pSgdma->identifier 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pSgdma->identifier)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pSgdma->firstDescLo 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pSgdma->firstDescLo)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pSgdma->firstDescHi 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pSgdma->firstDescHi)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pSgdma->firstDescAdj 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pSgdma->firstDescAdj)));
	dev_dbg(&(pDev->pci_devp->dev), "%s: pSgdma->DescCredits 0x%x\n", __FUNCTION__,ioread32((void *)&(pDev->XDmaRegs.pSgdma->DescCredits)));

	uiCompleted = pDev->pXDmaWBVirtAddr->uiCompletedDescCount;
    if (uiCompleted & XDMA_WB_ERR_MASK)
    {
    	pr_err("%s: DMA Error 0x%lx, count= 0x%lx (0x%lx)\n", __FUNCTION__,(uiCompleted & 0x7F000000) >> 24, uiCompleted & XDMA_WB_COUNT_MASK, uiCompleted);
    }
    else
    {
    	pr_err("%s: NO DMA Error 0x%lx, count= 0x%lx (0x%lx)\n", __FUNCTION__,(uiCompleted & 0x7F000000) >> 24, uiCompleted & XDMA_WB_COUNT_MASK, uiCompleted);
    }


}

int32_t dma_build_contiguous_sgdescriptor_list(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	int32_t iResult;
	int32_t iNumDescriptors;
	DMA_DESCRIPTOR_LIST *pCurrent;
	struct scatterlist *pSG;
	struct sg_table *pSGTable;
	uint64_t uiLastDmaAddress;
	uint64_t uiLastDmaLength;
	uint64_t uiTotalDmaLength;
	int i;

	if ((NULL == pDev) || (NULL == pSGMap))
		return(-EFAULT);

	dev_dbg(&(pDev->pci_devp->dev), "%s: %s: ENTERED\n", DEVICE_NAME, __FUNCTION__);

	// Free up old stuff
	dma_free_contiguous_sgdescriptor_list(pDev, pSGMap);

	iResult= 0;
	iNumDescriptors= 0;
	if (pSGMap->sgTable.nents <= 0)
		return(-EFAULT);
	pSGTable = &(pSGMap->sgTable);
	if (NULL == pSGTable)
	{
		dev_dbg(&(pDev->pci_devp->dev), "%s: NULL == pSGTable\n", __FUNCTION__);
		return(-EFAULT);
	}
	pSG= pSGMap->sgTable.sgl;
	if (NULL == pSG)
	{
		dev_dbg(&(pDev->pci_devp->dev), "%s: NULL == pSG\n", __FUNCTION__);
		return(-EFAULT);
	}

	dev_dbg(&(pDev->pci_devp->dev), "%s: First page_link= 0x%lx, offset= 0x%x, length= 0x%x, dma_address= 0x%llx\n",
									 __FUNCTION__,pSG->page_link, pSG->offset,sg_dma_len(pSG),sg_dma_address(pSG));

	// Create and init the first descriptor
	pSGMap->pContigDescList= kcalloc(1, sizeof(DMA_DESCRIPTOR_LIST), GFP_KERNEL);
	if (pSGMap->pContigDescList)
	{
		pCurrent= pSGMap->pContigDescList;
		uiLastDmaAddress= sg_dma_address(pSG);
		uiLastDmaLength= sg_dma_len(pSG);
		uiTotalDmaLength= uiLastDmaLength;
		pCurrent->dmaEngineDescriptor.numBytes= cpu_to_le32(uiTotalDmaLength);
		pCurrent->dmaEngineDescriptor.dstAddrLo= cpu_to_le32(PCI_DMA_LO(uiLastDmaAddress));
		pCurrent->dmaEngineDescriptor.dstAddrHi= cpu_to_le32(PCI_DMA_HI(uiLastDmaAddress));
		pCurrent->pNext= NULL;
		pSG= sg_next(pSG);
		for (i=1; (i<pSGMap->sgTable.nents) && (iResult >= 0); ++i, pSG= sg_next(pSG))
		{
			// if the total length of the previous descriptors and this descriptor is larger than
			// what the hardware will allow, or it is not contiguous with the next descriptor
			// then create a new descriptor
			if ( ((uiTotalDmaLength + sg_dma_len(pSG)) > XDMA_DESC_MAX_BYTE_LEN) || (sg_dma_address(pSG) != (uiLastDmaAddress + uiLastDmaLength)) )
			//if ( ((uiTotalDmaLength + sg_dma_len(pSG)) > (1024*1024)) || (sg_dma_address(pSG) != (uiLastDmaAddress + uiLastDmaLength)) )
			{
				// Need a new descriptor.....First update the length of the
				// current descriptor
				pCurrent->dmaEngineDescriptor.numBytes= cpu_to_le32(uiTotalDmaLength);
				iNumDescriptors++;

				if (iNumDescriptors < 10)
				{
					dev_dbg(&(pDev->pci_devp->dev), "%s: Finalized %dth descriptor: length: 0x%x, 0x%08x%08x\n",__FUNCTION__,iNumDescriptors,
							le32_to_cpu(pCurrent->dmaEngineDescriptor.numBytes),
							le32_to_cpu(pCurrent->dmaEngineDescriptor.dstAddrHi),
							le32_to_cpu(pCurrent->dmaEngineDescriptor.dstAddrLo));
				}

				// Now allocate and initialize a new one
				pCurrent->pNext = kcalloc(1, sizeof(DMA_DESCRIPTOR_LIST), GFP_KERNEL);
				if (pCurrent->pNext)
				{
					pCurrent= pCurrent->pNext;
					uiLastDmaAddress= sg_dma_address(pSG);
					uiLastDmaLength= sg_dma_len(pSG);
					uiTotalDmaLength= uiLastDmaLength;
					pCurrent->dmaEngineDescriptor.numBytes= cpu_to_le32(uiTotalDmaLength);
					pCurrent->dmaEngineDescriptor.dstAddrLo= cpu_to_le32(PCI_DMA_LO(uiLastDmaAddress));
					pCurrent->dmaEngineDescriptor.dstAddrHi= cpu_to_le32(PCI_DMA_HI(uiLastDmaAddress));
					pCurrent->pNext= NULL;
				}
				else
				{
					pr_err("%s:%s kcalloc error allocating descriptor\n", DEVICE_NAME, __FUNCTION__);
					iResult = -ENOMEM;
				}
			}
			else
			{
				// still building this descriptor, so just update the total length
				// and last pointers....
				uiLastDmaAddress= sg_dma_address(pSG);
				uiLastDmaLength= sg_dma_len(pSG);
				uiTotalDmaLength += sg_dma_len(pSG);
			}
		}

		// if all is well, Finalize the last descriptor
		if (iNumDescriptors >= 0)
		{
			pCurrent->dmaEngineDescriptor.numBytes= cpu_to_le32(uiTotalDmaLength);
			iNumDescriptors++;
			pSGMap->uiNumContigDescriptors= iNumDescriptors;

			dev_dbg(&(pDev->pci_devp->dev), "%s: Finalized %dth (last) descriptor: length: 0x%x, 0x%08x%08x\n",__FUNCTION__,iNumDescriptors,
					le32_to_cpu(pCurrent->dmaEngineDescriptor.numBytes),
					le32_to_cpu(pCurrent->dmaEngineDescriptor.dstAddrHi),
					le32_to_cpu(pCurrent->dmaEngineDescriptor.dstAddrLo));

			// Now create the Hardware Descriptors that will be loaded onto the FPGA
			iResult= create_hw_descriptors(pDev, pSGMap, pSGMap->pContigDescList, iNumDescriptors);

			DumpDescriptors(pDev, pSGMap);
		}
	}
	else
	{
		pr_err("%s:%s kcalloc error allocating initial descriptor\n", DEVICE_NAME, __FUNCTION__);
		iResult = -ENOMEM;
	}

	dev_dbg(&(pDev->pci_devp->dev), "%s: returning %d, %d descriptors\n",__FUNCTION__,iResult, pSGMap->uiNumContigDescriptors);

	return(iResult);
}

void dma_free_contiguous_sgdescriptor_list(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	DMA_DESCRIPTOR_LIST *pList;
	DMA_DESCRIPTOR_LIST *pNext;

	if ((NULL == pDev) || (NULL == pSGMap))
		return;

	// Free the descriptor DMA coherent pages
	free_desc_pages(pDev, pSGMap);

	// Free the contiguous SG list
	pList= pSGMap->pContigDescList;
	while (pList)
	{
		pNext= pList->pNext;
	    kfree(pList);
	    pList= pNext;
	}
	pSGMap->uiNumContigDescriptors= 0;
	pSGMap->pContigDescList= NULL;
}


//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
/// Static Function IMplementations
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
int32_t create_hw_descriptors(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap, DMA_DESCRIPTOR_LIST *pList, int32_t iNumDescriptors)
{
	int32_t iResult;
	DMA_DESCRIPTOR *pCurrDescriptor;
	DMA_DESCRIPTOR *pPrevDescriptor;
	uint32_t uiNumPages;
	uint32_t uiNumDescInPage;
	uint32_t uiPage;
	uint64_t uiNextPhysAddr;
	int32_t  i;

	iResult= 0;

	// Figure out how many 4K pages/address boundaries we need
	// for all the descriptors.  The max numbers of descriptors we can have
	// is only 63 because there are only 6 bits in the Nxt_adj field of the
	// control word in the descriptor.
	uiNumPages= (iNumDescriptors / XDMA_DESC_MAX_NXTADJ_NUM) + 1;

	// Allocate the descriptors
	// A set of descriptors cannot cross a 4K boundary (Xilinx XDMA doc - chapter 2 Descriptors
	// We allocate a page per set of 64 descriptors + 1 more page in case the first page is
	// no aligned to a 4 K boundary.  The XiLinx doc does only say a set of 64 can not cross
	// a 4K boundary, so we could have multiple sets of 64 within a 4K boundary (I think)
	// But for now, we will put each set in it's own page that starts on a 4K boundary and perhaps
	// make that optimization later after things are working.  From what I have seen so far, we will
	// only ever need 1 pages for the descriptors anyways.
	uiNumPages++;
	iResult= alloc_desc_pages(pDev, uiNumPages, &pSGMap->descInfo);
	if (iResult >= 0)
	{
		pSGMap->pHWDescriptors= pSGMap->descInfo.pDescVirtp;
		pPrevDescriptor= NULL;

		// all is well so far- we got space for the descriptors
		// Now fill them in......max 64 to a page....
		uiNumDescInPage= XDMA_DESC_MAX_NXTADJ_NUM;
		uiPage= 0;
		while (iNumDescriptors > 0)
		{
			// Figure out how many descriptors will be in this page
			if (iNumDescriptors < XDMA_DESC_MAX_NXTADJ_NUM)
				uiNumDescInPage= iNumDescriptors;

			uiNextPhysAddr= pSGMap->descInfo.uiDescPhysp + (uiPage  * XDMA_DESC_4K_SIZE);
			pCurrDescriptor= (DMA_DESCRIPTOR *)((uint8_t *)pSGMap->pHWDescriptors + (uiPage * XDMA_DESC_4K_SIZE));

			// update the previous descriptor appropriately
			if (pPrevDescriptor)
			{
				pPrevDescriptor->nextLo= cpu_to_le32(PCI_DMA_LO(uiNextPhysAddr));
				pPrevDescriptor->nextHi= cpu_to_le32(PCI_DMA_HI(uiNextPhysAddr));
				pCurrDescriptor->control|= uiNumDescInPage << XDMA_DESC_CONTROL_NUM_ADJ_SHIFT;
			}


			// The next_adj field is messed up... It starts at uiNumDescInPage-2 and the
			// second to last descriptor has it set as 0.  The last descriptor
			// has the number of descriptors in the next set.
			for (i=(int32_t)(uiNumDescInPage-1); i>=0; --i)
			{
				pCurrDescriptor->control= XDMA_DESC_CONTROL_MAGIC;
				if (i > 0)
					pCurrDescriptor->control|= (i - 1) << XDMA_DESC_CONTROL_NUM_ADJ_SHIFT;
				else
					pCurrDescriptor->control|= i << XDMA_DESC_CONTROL_NUM_ADJ_SHIFT;
				pCurrDescriptor->control|= XDMA_DESC_CONTROL_COMPLETED_BIT | XDMA_DESC_CONTROL_EOP_BIT;
				pCurrDescriptor->numBytes= pList->dmaEngineDescriptor.numBytes;
				pCurrDescriptor->dstAddrLo= pList->dmaEngineDescriptor.dstAddrLo;
				pCurrDescriptor->dstAddrHi= pList->dmaEngineDescriptor.dstAddrHi;
				pCurrDescriptor->srcAddrLo= 0;
				pCurrDescriptor->srcAddrHi= 0;

				uiNextPhysAddr+= sizeof(DMA_DESCRIPTOR);
				pCurrDescriptor->nextLo= cpu_to_le32(PCI_DMA_LO(uiNextPhysAddr));
				pCurrDescriptor->nextHi= cpu_to_le32(PCI_DMA_HI(uiNextPhysAddr));

				pPrevDescriptor= pCurrDescriptor;
				pCurrDescriptor++;
				iNumDescriptors--;
				pList= pList->pNext;
			}
		}

		// Update the last descriptor so things stop- here is where
		// we will set it back to the first descriptor when we make that
		// optimization.....
		if (pPrevDescriptor)
		{
			pPrevDescriptor->nextLo= cpu_to_le32(PCI_DMA_LO(0));
			pPrevDescriptor->nextHi= cpu_to_le32(PCI_DMA_HI(0));
			pPrevDescriptor->control|= XDMA_DESC_CONTROL_STOP_BIT;
		}
	}
	else
	{
		pSGMap->pHWDescriptors= NULL;
	}

	return(iResult);
}

int32_t alloc_desc_pages(fli_cam_data_t *pDev, uint32_t uiNumPages, DMA_DESC_INFO *pDescInfo)
{
	int32_t    iResult = 0;

	if (NULL == pDescInfo)
		return(-ENOMEM);

	dev_dbg(&(pDev->pci_devp->dev), "Entry %s \n", __FUNCTION__);

	pDescInfo->uiAllocationSize = uiNumPages * XDMA_DESC_4K_SIZE;

	pDescInfo->pVirtStart = (DMA_DESCRIPTOR *)dma_alloc_coherent(&(pDev->pci_devp->dev),
			                                                      pDescInfo->uiAllocationSize,
																  &(pDescInfo->uiPhysStart),
																  GFP_KERNEL);
	if (pDescInfo->pVirtStart)
	{
		if ((unsigned long)pDescInfo->uiPhysStart & XDMA_DESC_4K_MASK)
		{
			// This is unexpected, the buffer does not start on a 4K boundary- which is
			// why we allocated an extra one.  So go the next 4K boundary
			pDescInfo->pDescVirtp= pDescInfo->pVirtStart + (XDMA_DESC_4K_SIZE - ((uint64_t)pDescInfo->pVirtStart & XDMA_DESC_4K_MASK));
			pDescInfo->uiDescPhysp= pDescInfo->uiPhysStart + (XDMA_DESC_4K_SIZE - (pDescInfo->uiPhysStart & XDMA_DESC_4K_MASK));
		}
		else
		{
			// All is as expected- the allocation starts on a page boundary
			// so w just use the start
			pDescInfo->pDescVirtp= pDescInfo->pVirtStart;
			pDescInfo->uiDescPhysp= pDescInfo->uiPhysStart;
		}
	}
	else
	{
		pDescInfo->uiAllocationSize= 0;
		pr_err("Error allocating dma descriptor buffer (%s)\n", DEVICE_NAME);
		iResult = -ENOMEM;
	}

   dev_dbg(&(pDev->pci_devp->dev), "%s: Leaving (%d)\n", __FUNCTION__, iResult);

   return(iResult);

}

void free_desc_pages (fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
   if (NULL == pDev)
	   return;

   //dev_dbg(&(this_devp->pci_devp->dev), "Entry %s \n", __FUNCTION__);

   if (pSGMap->descInfo.pVirtStart)
   {
      dma_free_coherent (&(pDev->pci_devp->dev), pSGMap->descInfo.uiAllocationSize, pSGMap->descInfo.pVirtStart, pSGMap->descInfo.uiPhysStart);
      pSGMap->descInfo.pVirtStart = NULL;
      pSGMap->descInfo.uiAllocationSize= 0;
   }

   //dev_dbg(&(this_devp->pci_devp->dev), "Exit %s \n", __FUNCTION__);
}


void DumpDescriptors(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	DMA_DESCRIPTOR_LIST *pList;
	DMA_DESCRIPTOR      *pHWDesc;
	uint32_t            uiIsLastDescriptor;
	uint64_t            uiNextAddr;
	uint32_t            uiPage;
	uint32_t            uiSafetyCount;

	// Dump the descriptor info
	if ((NULL == pDev) || (NULL == pSGMap))
		return;

	// dump the contiguous SG list
	dev_dbg(&(pDev->pci_devp->dev), "%s: Contiguous Descriptors Info (%d descriptors)\n", __FUNCTION__, pSGMap->uiNumContigDescriptors);
	pList= pSGMap->pContigDescList;
	while (pList)
	{
		dev_dbg(&(pDev->pci_devp->dev), "    control: 0x%x, numBytes= 0x%x, srcAddr= 0x%08x%08x, destAddr= 0x%08x%08x, nextAddr= 0x%08x%08x\n",
				pList->dmaEngineDescriptor.control,pList->dmaEngineDescriptor.numBytes,pList->dmaEngineDescriptor.srcAddrHi, pList->dmaEngineDescriptor.srcAddrLo,
				pList->dmaEngineDescriptor.dstAddrHi,pList->dmaEngineDescriptor.dstAddrLo,
				pList->dmaEngineDescriptor.nextHi, pList->dmaEngineDescriptor.nextLo);

	    pList= pList->pNext;
	}

	// Dump the hardware Info
	dev_dbg(&(pDev->pci_devp->dev), "%s: Hardware Descriptor Info\n", __FUNCTION__);
	dev_dbg(&(pDev->pci_devp->dev), "    uiAllocationSize= %d bytes, pVirtStart= 0x%p, uiPhysStart= 0x%llx, pDescVirtp= 0x%p, uiDescPhysp= 0x%llx\n",
			pSGMap->descInfo.uiAllocationSize, pSGMap->descInfo.pVirtStart, pSGMap->descInfo.uiPhysStart,
			pSGMap->descInfo.pDescVirtp, pSGMap->descInfo.uiDescPhysp);

	// Dump the Hardware descriptors
	pHWDesc= pSGMap->descInfo.pDescVirtp;
	dev_dbg(&(pDev->pci_devp->dev), "%s: DMA Coherent Descriptors at pDescVirtp= 0x%p\n", __FUNCTION__,pHWDesc);
	uiPage= 0;
	uiSafetyCount= 0;
	uiIsLastDescriptor= 0;
	while ((pHWDesc) && (uiSafetyCount < 64))
	{
		dev_dbg(&(pDev->pci_devp->dev), "    control: 0x%x (adj=%d), length= 0x%x, src= 0x%08x%08x, dest= 0x%08x%08x, next= 0x%08x%08x\n",
				pHWDesc->control,XDMA_DESC_CONTROL_GET_ADJ_DESC_NUM(pHWDesc->control), pHWDesc->numBytes,
				pHWDesc->srcAddrHi, pHWDesc->srcAddrLo,
				pHWDesc->dstAddrHi,pHWDesc->dstAddrLo,
				pHWDesc->nextHi, pHWDesc->nextLo);

		// is this the last descriptor- the stop bit will be set
		if (0 == (pHWDesc->control & XDMA_DESC_CONTROL_STOP_BIT))
		{
			// Are we on the last descriptor?
			if (uiIsLastDescriptor > 0)
			{
				// see if there are more on another page
				uiNextAddr= ((uint64_t)pHWDesc->nextHi << 32) | ((uint64_t)pHWDesc->nextLo);
				if (uiNextAddr > 0)
				{
					uiIsLastDescriptor= 0;
					pHWDesc= (DMA_DESCRIPTOR *)((uint8_t*)pSGMap->descInfo.pDescVirtp + (uiPage * XDMA_DESC_4K_SIZE));
				}
				else
				{
					// all done
					pHWDesc= NULL;
				}
			}
			else
			{
				// the second to last descriptor has the nextAdj field set to 0
				// So the next time through here, we will be on the last descriptor
				if (0 == XDMA_DESC_CONTROL_GET_ADJ_DESC_NUM(pHWDesc->control))
					uiIsLastDescriptor= 1;

				// just get the next adjacent descriptor
				pHWDesc++;
			}
		}
		else
		{
			// all done- stop bit set-
			pHWDesc= NULL;

		}

		uiSafetyCount++;
	}
}

void clear_poll_buffers(volatile XDMA_POLL_WB *pPollBuff)
{
	if(NULL == pPollBuff)
		return;
   memset((void *)pPollBuff, 0, sizeof(XDMA_POLL_WB));
}

