/*
 * dma_engine_types.h
 *
 *  Created on: Aug 25, 2020
 *      Author: tom
 */

#ifndef DMA_ENGINE_TYPES__H_
#define DMA_ENGINE_TYPES__H_

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/types.h>
#include <linux/device.h>

#include "flipcie_regblocks.h"


// XDMA Scatter/Gather DMA Descriptor for a single contiguous memory block transfer.
// The descriptors are in root complex memory, and the bytes in the 32-bit
// words must be in little-endian byte ordering.
#pragma pack(1)
typedef struct
{
    uint32_t control;
    uint32_t numBytes;   // transfer length in bytes
    uint32_t srcAddrLo;  // source or C2H stream metadata address (low 32-bit)
    uint32_t srcAddrHi;  // source or C2H stream metadata address (high 32-bit)
    uint32_t dstAddrLo;  // destination address (low 32-bit)
    uint32_t dstAddrHi;  // destination address (high 32-bit)
                         // next descriptor in the single-linked list of descriptors,
                         // this is the bus address of the next descriptor in the root complex memory.
    uint32_t nextLo;     // next desc address (low 32-bit)
    uint32_t nextHi;     // next desc address (high 32-bit)
} DMA_DESCRIPTOR;
#pragma pack()

// List of DMA Descriptors
typedef struct _dma_desc_list_t
{
	DMA_DESCRIPTOR dmaEngineDescriptor;
	struct _dma_desc_list_t *pNext;

} DMA_DESCRIPTOR_LIST;

typedef struct _dma_descriptor_info_t
{
	uint32_t uiAllocationSize;
	void *pVirtStart;         /// Where this allocation starts in virtual memory
	dma_addr_t uiPhysStart;   /// Where this allocation starts in physical memory

	DMA_DESCRIPTOR *pDescVirtp;   ///< Virtual pointer to descriptors within this allocation
	dma_addr_t      uiDescPhysp;  ///< physical pointer to descriptors within this allocation
} DMA_DESC_INFO;


#define XDMA_DESC_MAX_BYTE_LEN           ((1 << 28) -1)     ///< Max byte length of single DMA transfer
#define XDMA_DESC_MAX_NXTADJ_NUM         ((1 << 6))         ///< Max number of adjacent descriptors (Nxt_adj field in control)
                                                            ///  can have 64, but 63 Nxt_adj can follow th4 first
#define XDMA_DESC_4K_SIZE                (1 << 12)
#define XDMA_DESC_4K_MASK                (0xFFF)

// Descriptor related macros and defines
#define XDMA_DESC_CONTROL_MAGIC                  (0xAD4B0000)
#define XDMA_DESC_CONTROL_NUM_ADJ_SHIFT          (8)
#define XDMA_DESC_CONTROL_COMPLETED_BIT          (1 << 1)
#define XDMA_DESC_CONTROL_EOP_BIT                (1 << 4)
#define XDMA_DESC_CONTROL_STOP_BIT               (1 << 0)
#define XDMA_DESC_CONTROL_GET_ADJ_DESC_NUM(__xctrl)       ((__xctrl >> XDMA_DESC_CONTROL_NUM_ADJ_SHIFT) & (0x3F))


/* obtain the 32 most significant (high) bits of a 32-bit or 64-bit address */
#define PCI_DMA_HI(addr) ((addr >> 16) >> 16)
/* obtain the 32 least significant (low) bits of a 32-bit or 64-bit address */
#define PCI_DMA_LO(addr) (addr & 0xffffffffUL)


#endif /* DMA_ENGINE_TYPES_H_ */
