//***************************************************************
/// @file flipcie_Cam.c
/// @brief register read/write handling for the FLI Host card device driver
/// @author      ttatakis
/// @date
/// @copyright
///
//
//
//***************************************************************

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/types.h>
#include <linux/device.h>
//#include <linux/delay.h>
//#include <linux/kthread.h>

//#include <asm/byteorder.h>
//#include <asm/cacheflush.h>

#include "flipcie_priv.h"
#include "flipcie_ioctl.h"
#include "flipcie_dma.h"
#include "flipcie_regs.h"
#include "dma_engine.h"


/// @name Driver Version Constants
/// @{
#define FLI_CAM_MAJOR_VERSION (1)
#define FLI_CAM_MINOR_VERSION (0)
#define FLI_CAM_BUILD_NUMBER  (0)
///@}



typedef enum
{
   FLI_REG_ACCESS_READ,
   FLI_REG_ACCESS_WRITE
} FLI_REG_ACCESS_TYPE;


fli_cam_data_t *s_pFLI_cam_data[NUM_FLI_PCIE_DEVICE];
unsigned int   s_MessagBuffer[FLI_IOCTL_MESSAGE_LENGTH_MAX];

#define FLI_IOCTL_REF_FIFO_MAX (1024)
unsigned int   s_uiReferenceFifoBuffer[FLI_IOCTL_REF_FIFO_MAX];


static int fli_cam_open(struct inode *inode, struct file *filp);
static int fli_cam_release(struct inode *inode, struct file *filp);
static long fli_cam_ioctl(struct file *filp, unsigned int cmd, unsigned long arg);


static int  BufferToStreamInit(FLI_CAM_IOCTL_FRAME_READ_T  *pBuffer, FLI_CAM_IOCTL_STREAM_INIT_T *pStreamInit);
static int  CopyStreamInitFromUser(FLI_CAM_IOCTL_STREAM_INIT_T *pUser,FLI_CAM_IOCTL_STREAM_INIT_T *pKernel);
static void FreeStreamInit(FLI_CAM_IOCTL_STREAM_INIT_T *pStreamInit);

static long DoRegReadWrite(struct file *filp,
                          unsigned int cmd,
                          unsigned long arg,
                          FLI_REG_ACCESS_TYPE access_type);
static long FillVersionInfo(struct file *filp, DRV_INFO *pInfo);
static fli_sg_mapping_t *FindDMAMapping(fli_cam_data_t *pDev, unsigned char *pBuffer);
static long DoRegRead(fli_cam_data_t *pDev, void * pBar, long lReg, uint32_t *pValue);
static long DoRegWrite(fli_cam_data_t *pDev, void * pBar, long lReg, uint32_t uiValue);
static int  DataTxFifoEmpty(fli_cam_data_t *pDev);
static int  DataRxMsgReceived(fli_cam_data_t *pDev, uint32_t *pNumDwords);
static long DoSendMessage(struct file *filp, unsigned long arg);
static long DoGetResponse(struct file *filp, unsigned long arg);
static long DoFrameRead(struct file *filp, unsigned long arg);
static long DoStreamRead(struct file *filp, unsigned long arg);
static long DoStreamInit(struct file *filp, unsigned long arg);
static long DoStreamDeinit(struct file *filp);
static long DoWriteReferenceFifo(struct file *filp, unsigned long arg);
static void DumpDMAMappings(fli_cam_data_t *pDev);
static void DumpConfig(struct pci_dev *pDev);

/// Mapping of standard file operations to driver functions
struct file_operations fli_cam_fops =
{
   owner: THIS_MODULE,
   open: fli_cam_open,
   //llseek: fli_cam_llseek,
   //read: fli_cam_read,
   //write: fli_cam_write,
   release: fli_cam_release,
   unlocked_ioctl: fli_cam_ioctl
};

/// Driver interface for when a device file is opened
static int fli_cam_open(struct inode *inode, struct file *filp)
{
   bool           found;
   int            idx;
   int ret = 0;

   pr_debug("open (%s)\n", DEVICE_NAME);

   // Find matching device data, data was initialized in probe function
   found = false;
   for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
   {
      // Find the data instance that matches this device numbers
      if ((s_pFLI_cam_data[idx]->minor == iminor(inode))
            && (s_pFLI_cam_data[idx]->major == imajor(inode)))
      {
         // Save device data, accessible from the file data
         filp->private_data = s_pFLI_cam_data[idx];
         found = true;
      }
   }

    if (!found)
    {
      pr_err("Error device data not found %d %d (%s)\n", iminor(inode), imajor(inode), DEVICE_NAME);
      ret = -ENODEV;
    }

   return ret;
} // fli_cam_open

/// Driver interface for when a device file is closed
static int fli_cam_release(struct inode *inode, struct file *filp)
{
   pr_debug("%s: release\n", __FUNCTION__);
   return 0;
} // fli_cam_release

/// Driver interface for a file ioctl action
static long fli_cam_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
   long           ret = 0;
   fli_cam_data_t *pDev;  // pointer to per device memory
   //fli_cam_IOCTL_RW_REG_T k_rw;

   //pr_debug("%s: ioctl\n", DEVICE_NAME);

   pDev = (fli_cam_data_t *)(filp->private_data);

   switch (cmd)
   {
		// non-dma read
		case FLI_CAM_IOCTL_READ_REG:
			//dev_dbg(&(this_devp->pci_devp->dev), "%s: IOCTL_READ_REG\n", __FUNCTION__);
			ret = DoRegReadWrite(filp, cmd, arg, FLI_REG_ACCESS_READ);
			break;

		// non-dma write
		case FLI_CAM_IOCTL_WRITE_REG:
			//dev_dbg(&(this_devp->pci_devp->dev), "%s: IOCTL_WRITE_REG\n", __FUNCTION__);
			ret = DoRegReadWrite(filp, cmd, arg, FLI_REG_ACCESS_WRITE);
			break;

		// non-dma write
		case FLI_CAM_IOCTL_SEND_MESSAGE:
			//dev_dbg(&(this_devp->pci_devp->dev), "%s: IOCTL_SEND_MESSAGE\n", __FUNCTION__);
			ret = DoSendMessage(filp, arg);
			break;
		case FLI_CAM_IOCTL_GET_RESPONSE:
			//dev_dbg(&(this_devp->pci_devp->dev), "%s: IOCTL_GET_RESPONSE\n", __FUNCTION__);
			ret = DoGetResponse(filp, arg);
			break;

		case FLI_CAM_IOCTL_FRAME_READ:
			//dev_dbg(&(pDev->pci_devp->dev), "%s: IOCTL_FRAME_READ\n", __FUNCTION__);
			ret = DoFrameRead(filp, arg);
			break;

		case FLI_CAM_IOCTL_STREAM_INIT:
			dev_dbg(&(pDev->pci_devp->dev), "%s: FLI_CAM_IOCTL_STREAM_INIT\n", __FUNCTION__);
			ret = DoStreamInit(filp, arg);
			break;

		case FLI_CAM_IOCTL_STREAM_DEINIT:
			dev_dbg(&(pDev->pci_devp->dev), "%s: FLI_CAM_IOCTL_STREAM_DEINIT\n", __FUNCTION__);
			ret = DoStreamDeinit(filp);
			break;

		case FLI_CAM_IOCTL_STREAM_READ:
			//dev_dbg(&(pDev->pci_devp->dev), "%s: IOCTL_STREAM_READ\n", __FUNCTION__);
			ret = DoStreamRead(filp, arg);
			break;

		case FLI_CAM_IOCTL_WRITE_REF_FIFO:
			ret = DoWriteReferenceFifo(filp, arg);
			break;

		case FLI_CAM_IOCTL_VERSION:
			dev_dbg(&(pDev->pci_devp->dev), "%s: IOCTL_VERSION\n", __FUNCTION__);
			//ret = fli_cam_MAKE_VERSION(FLI_PCIE_MAJOR_VERSION, FLI_PCIE_MINOR_VERSION, FLI_PCIE_BUILD_NUMBER);
			ret= FillVersionInfo(filp,(DRV_INFO *)arg);
			break;

		default:
			pr_err("%s: unknown ioctl command %d\n", DEVICE_NAME, cmd);
			ret = -ENOSYS;
			break;
	}

   return ret;
}


// Called by by the parent driver when a device is discovered
int fli_cam_probe(struct pci_dev *pPciDev, struct class *fli_class, int idx, int major, int minor)
{
   int             ret;
   fli_cam_data_t * pCamDev;
   dev_t           devno;
   struct device * devicep;

   dev_dbg(&(pPciDev->dev), "%s: %s\n", DEVICE_NAME,__FUNCTION__);

   DumpConfig(pPciDev);

   // if device data is available
   if (s_pFLI_cam_data[idx]->pci_devp == NULL)
   {
	   pCamDev = s_pFLI_cam_data[idx];
	   pCamDev->pci_devp = pPciDev;

      // Initialize major and minor numbers to use as devices are found
	   pCamDev->major = major;
	   pCamDev->minor = minor;

	   pCamDev->bar0.start = pci_resource_start(pPciDev, BAR_0 );
	   pCamDev->bar0.len   = pci_resource_len(pPciDev, BAR_0 );
	   pCamDev->bar0.flags = pci_resource_flags(pPciDev, BAR_0 );

      dev_dbg(&(pPciDev->dev), "Device info Registers: start 0x%lX, len 0x%lX, flags 0x%lX (%s)\n",
    		  pCamDev->bar0.start, pCamDev->bar0.len, pCamDev->bar0.flags, DEVICE_NAME);

      // Get a handle to the memory on the device
      pCamDev->bar0.hdl = pci_iomap(pPciDev, BAR_0, 0);
      dev_dbg(&(pPciDev->dev), "this_devp->bar0.hdl 0x%LX (%s)\n", (uint64_t)pCamDev->bar0.hdl, DEVICE_NAME);

      ret= flidma_probe(pPciDev, pCamDev);
      if (ret < 0)
      {
          pr_err("%s: %s: flidma_probe FAILED %d\n",DEVICE_NAME, __FUNCTION__,  ret);
          return ret;
      }

      // Set up for file operations
      cdev_init(&(pCamDev->cdev), &fli_cam_fops);
      pCamDev->cdev.owner = THIS_MODULE;

      // regenerate the device number
      devno = MKDEV(pCamDev->major, pCamDev->minor);
      printk("%s: major= %d, minor= %d\n", __FUNCTION__,pCamDev->major, pCamDev->minor);

      // Apply major and minor numbers
      ret = cdev_add(&(pCamDev->cdev), devno, 1);
      if (ret)
      {
         pr_err("%s: Fail cdev_add\n", DEVICE_NAME);
         return ret;
      }

      devicep = device_create(fli_class, NULL, devno, NULL, DEVICE_NAME "%d", idx);
      if ( IS_ERR(devicep))
      {
         pr_err("%s: Fail device_create %ld %d\n", DEVICE_NAME, (long)devicep, idx);
         return -ENOMEM;
      }
   }
   else
   {
      pr_err("%s: Error no device data avail\n", DEVICE_NAME);
      return -ENOMEM;
   }

   return 0;
} // fli_cam_probe

void fli_cam_remove(struct pci_dev *pPciDev, struct class *fli_class, int idx)
{
   fli_cam_data_t * pCamDev;
   dev_t           devno;

   pr_debug("%s: fli_cam_remove %d\n", DEVICE_NAME, idx);

   pCamDev = s_pFLI_cam_data[idx];

   if (flidma_remove(pPciDev, pCamDev) < 0)
   {
       pr_err("%s: %s: flidma_remove FAILED\n",DEVICE_NAME, __FUNCTION__);
   }

   // regenerate the device number
   devno = MKDEV(pCamDev->major, pCamDev->minor);

   device_destroy(fli_class, devno);

   //dev_dbg(&(pci_devp->dev), "device number, major 0x%X, minor 0x%X (%s)\n", this_devp->major, this_devp->minor, DEVICE_NAME);

   cdev_del(&(pCamDev->cdev));

   pCamDev->pci_devp = NULL;

} // fli_cam_remove

int fli_cam_init(void)
{
	int   idx;

	printk("%s: fli_cam_init\n", DEVICE_NAME);

	for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
	   s_pFLI_cam_data[idx] = NULL;

	for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
	{
		// allocate device data
		s_pFLI_cam_data[idx] = kzalloc(sizeof(fli_cam_data_t), GFP_KERNEL);
		if (!s_pFLI_cam_data[idx])
		{
			pr_err("%s: Fail kmalloc device data %d\n", DEVICE_NAME, idx);
			return (-ENOMEM);
		}

		// Mark device data as available
		s_pFLI_cam_data[idx]->pci_devp = NULL;

		// Init the DMA Mapping mutex
		mutex_init(&s_pFLI_cam_data[idx]->mutexDMAMapping);
	}

	pr_debug("%s: Loaded\n", DEVICE_NAME);
	return(0);
} // fli_cam_init

void fli_cam_exit(void)
{
	int   idx;

	printk("%s: fli_cam_exit\n", DEVICE_NAME);

	for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
	{
		if (s_pFLI_cam_data[idx])
			kfree(s_pFLI_cam_data[idx]);
	}
} // fli_cam_exit

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
/// Driver Helper Functions
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
int BufferToStreamInit(FLI_CAM_IOCTL_FRAME_READ_T  *pBuffer, FLI_CAM_IOCTL_STREAM_INIT_T *pStreamInit)
{
	int iResult;

	if ((NULL == pBuffer) || (NULL == pStreamInit))
	{
		pr_err("%s: %s: NULL pointer pBuffer= 0x%p, pStreamInit= 0x%p\n", DEVICE_NAME, __FUNCTION__, pBuffer, pStreamInit);
		return(-1);
	}
	iResult= 0;
	pStreamInit->uiNumBuffers= 1;
	pStreamInit->uiLength= pBuffer->uiLength;
	pStreamInit->ppBuffers=  (unsigned char **)kcalloc(1, sizeof(unsigned char *), GFP_KERNEL);
	if (pStreamInit->ppBuffers)
	{
		*(pStreamInit->ppBuffers)= pBuffer->pBuffer;
	}
	else
	{
		pr_err("%s: %s: Error allocating initialization buffer.\n", DEVICE_NAME, __FUNCTION__);
		iResult= -1;
	}

	return(iResult);
}

int CopyStreamInitFromUser(FLI_CAM_IOCTL_STREAM_INIT_T *pUser,FLI_CAM_IOCTL_STREAM_INIT_T *pKernel)
{
	int iResult;
	unsigned char **ppKUserBuffers;

	// Copy the base structure
	iResult= copy_from_user((void *)(pKernel), (void *)pUser, sizeof(FLI_CAM_IOCTL_STREAM_INIT_T));
	if (0 == iResult)
	{
		pr_debug("%s: pKernel: uiLength= %d, uiNumbuffers= %d, ppBuffers= 0x%px\n", __FUNCTION__, pKernel->uiLength, pKernel->uiNumBuffers, pKernel->ppBuffers);
		// Now allocate and copy the buffer pointers
		ppKUserBuffers= pKernel->ppBuffers;
		pKernel->ppBuffers=  (unsigned char **)kcalloc(pKernel->uiNumBuffers, sizeof(unsigned char *), GFP_KERNEL);
		if (pKernel->ppBuffers)
		{
			iResult= copy_from_user((void *)(pKernel->ppBuffers), (void *)ppKUserBuffers, sizeof(unsigned char *) * pKernel->uiNumBuffers);
			pr_debug("%s: pKernel: uiLength= %d, uiNumbuffers= %d, ppBuffers= 0x%px\n", __FUNCTION__, pKernel->uiLength, pKernel->uiNumBuffers, pKernel->ppBuffers);
			if (0 == iResult)
			{
				int i;
				pr_debug("%s: Stuff Copied\n", __FUNCTION__);
				pr_debug("%s: pKernel: uiLength= %d, uiNumbuffers= %d, ppBuffers= 0x%px\n", __FUNCTION__, pKernel->uiLength, pKernel->uiNumBuffers, pKernel->ppBuffers);
				for (i=0; i<pKernel->uiNumBuffers; ++i)
				{
					pr_debug("%s:        ppBuffers[%d]= 0x%px\n", __FUNCTION__, i, pKernel->ppBuffers[i]);
				}
			}
			else
			{
				pr_err("%s: %s: Error %d deep copying buffer pointers from user space\n", DEVICE_NAME, __FUNCTION__, iResult);
			}
		}
	}
	else
	{
		pr_err("%s: %s: Error %d copying base stream init structure from user space\n", DEVICE_NAME, __FUNCTION__, iResult);
	}

	return(iResult);
}

void FreeStreamInit(FLI_CAM_IOCTL_STREAM_INIT_T *pStreamInit)
{
	if (NULL == pStreamInit)
			return;

	if (pStreamInit->ppBuffers)
	{
		kfree(pStreamInit->ppBuffers);
		pStreamInit->ppBuffers= NULL;
		pStreamInit->uiNumBuffers= 0;
		pStreamInit->uiLength=0;
	}

}

// Non-dma read or write -
static long DoRegReadWrite(struct file *filp,
                                unsigned int cmd,
                                unsigned long arg,
                                FLI_REG_ACCESS_TYPE access_type)
{
	long              ret = 0;
	void *            pBar;
	FLI_CAM_IOCTL_RW_REG_T regRW;
	fli_cam_data_t *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	//dev_dbg(&(this_devp->pci_devp->dev), "Entry %s (%s)\n", __FUNCTION__, DEVICE_NAME);

	if (arg)
	{
		if (copy_from_user((void *)(&regRW), (void *)(arg), sizeof(FLI_CAM_IOCTL_RW_REG_T)) == 0)
		{
			// The fpaga supports two memory regions (BAR - base address register)
			if (regRW.bar == 0)
			{
				pBar = pDev->bar0.hdl;
			    if (regRW.reg >= pDev->bar0.len)
			    {
					pr_err("%s: BAR 0 register 0x%x out of range\n", DEVICE_NAME, regRW.reg);
					ret = -EFAULT;
			    }
			}
			else if (regRW.bar == 1)
			{
				pBar = pDev->bar1.hdl;
			    if (regRW.reg >= pDev->bar1.len)
			    {
					pr_err("%s: BAR 1 register 0x%x out of range\n", DEVICE_NAME, regRW.reg);
					ret = -EFAULT;
			    }
			}
			else
			{
				pr_err("%s: Unexpected BAR value (0x%X) I/O RW\n", DEVICE_NAME, regRW.bar);
				ret = -EFAULT;
			}

			// If all is well we can do the read/write
			if (ret == 0)
			{
				if (access_type == FLI_REG_ACCESS_READ)
				{
					ret= DoRegRead(pDev, pBar, regRW.reg, &regRW.val);
					if (0 == ret)
					{
						if (copy_to_user((void *)(arg), (void *)(&regRW), sizeof(FLI_CAM_IOCTL_RW_REG_T)) != 0)
						{
							pr_err("%s: ioctl read (%d) copy_to_user error\n", DEVICE_NAME, cmd);
							ret = -EFAULT;
						}
					}
				}
				else // write
				{
					ret= DoRegWrite(pDev, pBar, regRW.reg, regRW.val);
				}
			}
		}
		else
		{
			pr_err("%s: ioctl read (%d) copy_from_user error\n", DEVICE_NAME, cmd);
		}
	}
	else
	{
		pr_err("%s: ioctl read (%d) without data\n", DEVICE_NAME, cmd);
		ret = -EINVAL;
	}

	return ret;
} // fli_cam_ioctl_rw

static int s_iLastReadReg= -1;
static uint32_t s_uiLastReadValue= 0;
long DoRegRead(fli_cam_data_t *pDev, void * pBar, long lReg, uint32_t *pValue)
{
	long              ret = 0;

	*pValue = ioread32(pBar + lReg);

	// Debug only....The if keeps the debug down when an app is polling looking for
	// responses to arrive from the camera.
	if ((s_iLastReadReg != lReg) || (s_uiLastReadValue != *pValue))
	{
		//dev_dbg(&(pDev->pci_devp->dev), "%s: read value 0x%X from reg 0x%lX\n",__FUNCTION__, *pValue, lReg);
		s_iLastReadReg= lReg;
		s_uiLastReadValue= *pValue;
	}

    return(ret);

}

long DoRegWrite(fli_cam_data_t *pDev, void * pBar, long lReg, uint32_t uiValue)
{
	long              ret = 0;

	iowrite32(uiValue, pBar + lReg);
	//dev_dbg(&(pDev->pci_devp->dev), "%s: write value 0x%X to reg 0x%lX\n",__FUNCTION__, uiValue, lReg);

    return(ret);

}

int DataTxFifoEmpty(fli_cam_data_t *pDev)
{
	unsigned int uiValue;
	int  iResult;

	iResult= 0;
	if (DoRegRead(pDev, pDev->bar0.hdl, REG_DATA_CTRL, &uiValue) >= 0)
	{
		if (uiValue & QSFP_DC_TX_FIFO_EMPTY)
			iResult= 1;
	}

	return(iResult);
}

int DataRxMsgReceived(fli_cam_data_t *pDev, uint32_t *pNumDwords)
{
	unsigned int uiValue;
	int  iResult;

	iResult= 0;
	if (DoRegRead(pDev, pDev->bar0.hdl, REG_DATA_CTRL, &uiValue) >= 0)
	{
		if (uiValue & QSFP_DC_RX_MSG_RCVD)
			iResult= 1;

		*pNumDwords= QSFP_DC_RX_WORD_COUNT(uiValue);
	}

	return(iResult);
}

long DoSendMessage(struct file *filp, unsigned long arg)
{
	long                ret = 0;
	unsigned int        uiNumDWords;
	unsigned int        uiPadDWords;
	unsigned int        *pData;
	FLI_CAM_IOCTL_MSG_T kIoctl;
	fli_cam_data_t *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	dev_dbg(&(pDev->pci_devp->dev), "%s: ENTERED\n", __FUNCTION__);
	if (arg)
	{
		// Make sure the fifo is empty
		if (DataTxFifoEmpty(pDev))
		{
			if (copy_from_user((void *)(&kIoctl), (void *)(arg), sizeof(FLI_CAM_IOCTL_MSG_T)) == 0)
			{
				if (kIoctl.length < FLI_IOCTL_MESSAGE_LENGTH_MAX)
				{
					// Copy the actual data to kernel space
					uiNumDWords= kIoctl.length;
					if (copy_from_user((void *)(s_MessagBuffer), (void *)(kIoctl.pData), uiNumDWords * 4) == 0)
					{
						uiPadDWords= 0;
						if (0 != (uiNumDWords % FLI_IOCTL_MESSAGE_MIN_DWORDS))
							uiPadDWords= FLI_IOCTL_MESSAGE_MIN_DWORDS - (uiNumDWords % FLI_IOCTL_MESSAGE_MIN_DWORDS);

						// TX the user data
						pData= &s_MessagBuffer[0];
						while (uiNumDWords-- > 0)
							DoRegWrite(pDev, pDev->bar0.hdl, REG_DATA_XMIT_FIFO,*pData++);

						// TX any pad words
						while (uiPadDWords-- > 0)
							DoRegWrite(pDev, pDev->bar0.hdl,REG_DATA_XMIT_FIFO,0);

						// Kick off the transfer
						DoRegWrite(pDev, pDev->bar0.hdl,REG_DATA_CTRL,QSFP_DC_TX_MSG_SEND);
						DoRegWrite(pDev, pDev->bar0.hdl,REG_DATA_CTRL,0);
					}
					else
					{
						pr_err("%s: %s: failure copying message data from user space\n", DEVICE_NAME, __FUNCTION__);
						ret = -EINVAL;
					}
				}
				else
				{
					pr_err("%s: %s: message data %d bytes too large, max allowed is %d\n", DEVICE_NAME, __FUNCTION__, kIoctl.length, FLI_IOCTL_MESSAGE_LENGTH_MAX);
					ret = -EINVAL;
				}
			}
			else
			{
				pr_err("%s: %s: failure copying message ioctl from user space\n", DEVICE_NAME, __FUNCTION__);
				ret = -EINVAL;
			}
		}
		else
		{
			pr_err("%s: %s: Data TX FIFO not empty, can not send message\n", DEVICE_NAME, __FUNCTION__);
			ret = -EINVAL;
		}
	}
	dev_dbg(&(pDev->pci_devp->dev), "%s: LEAVING (%ld)\n", __FUNCTION__, ret);

	return(ret);
}

long DoGetResponse(struct file *filp, unsigned long arg)
{
	long                ret = 0;
	unsigned int        uiDWordsToRead;
	unsigned int        uiDWordsToReadSave;
	unsigned int        uiDWordsTotalRead;
	unsigned int        uiDWordsLeft;
	unsigned int        *pData;
	FLI_CAM_IOCTL_MSG_T kIoctl;
	fli_cam_data_t *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	dev_dbg(&(pDev->pci_devp->dev), "%s: ENTERED\n", __FUNCTION__);
	if (arg)
	{
		// Make sure we have a response ready
		if (DataRxMsgReceived(pDev, &uiDWordsLeft))
		{
			if (copy_from_user((void *)(&kIoctl), (void *)(arg), sizeof(FLI_CAM_IOCTL_MSG_T)) == 0)
			{
				if (kIoctl.length >= uiDWordsLeft)
				{
					uiDWordsTotalRead= 0;
					while ((uiDWordsLeft > 0) && (ret >= 0))
					{
						uiDWordsToRead= uiDWordsLeft;
						uiDWordsToReadSave= uiDWordsLeft;
						pData= &s_MessagBuffer[0];
						if (uiDWordsToRead > FLI_IOCTL_MESSAGE_LENGTH_MAX)
							uiDWordsToRead= FLI_IOCTL_MESSAGE_LENGTH_MAX;
						while (uiDWordsToRead-- > 0)
							DoRegRead(pDev,pDev->bar0.hdl,REG_DATA_RCV_FIFO, pData++);

						// copy this chunk back to the user
						if (copy_to_user((void *)(&kIoctl.pData[uiDWordsTotalRead]), (void *)(s_MessagBuffer), uiDWordsToReadSave * 4) == 0)
						{
							uiDWordsLeft-= uiDWordsToReadSave;
							uiDWordsTotalRead+= uiDWordsToReadSave;
							ret= uiDWordsTotalRead;
						}
						else
						{
							pr_err("%s: %s: Error copying to user space, buffer size %d, copied %d, tried %d \n", DEVICE_NAME, __FUNCTION__, kIoctl.length, uiDWordsTotalRead, uiDWordsToReadSave );
							ret = -EINVAL;
						}
					}

				}
				else
				{
					pr_err("%s: %s: User buffer size %d too small, needed %d\n", DEVICE_NAME, __FUNCTION__, kIoctl.length, uiDWordsLeft);
					ret = -EINVAL;
				}
			}
			else
			{
				pr_err("%s: %s: failure copying message ioctl from user space\n", DEVICE_NAME, __FUNCTION__);
				ret = -EINVAL;
			}
		}
		else
		{
			pr_err("%s: %s: Data Rx FIFO empty, can not get response\n", DEVICE_NAME, __FUNCTION__);
			ret = -EINVAL;
		}
	}

	dev_dbg(&(pDev->pci_devp->dev), "%s: LEAVING (%ld)\n", __FUNCTION__, ret);

	return(ret);
}


long DoFrameRead(struct file *filp, unsigned long arg)
{
	long                        ret = 0;
	FLI_CAM_IOCTL_FRAME_READ_T  kIoctl;
	FLI_CAM_IOCTL_STREAM_INIT_T streamInit;
	fli_cam_data_t              *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	dev_dbg(&(pDev->pci_devp->dev), "%s: Entry\n", __FUNCTION__);
	if (arg)
	{
		if (copy_from_user((void *)(&kIoctl), (void *)(arg), sizeof(FLI_CAM_IOCTL_FRAME_READ_T)) == 0)
		{
			if (BufferToStreamInit(&kIoctl, &streamInit) >= 0)
			{
				// Create the DMA Descriptors
				if (flidma_create_descriptors(pDev, &streamInit) >= 0)
				{
					// Kick off the DMA
					flidma_start(pDev, &pDev->pStreamSgMapping[0]);

					// Wait for it to complete
					ret= flidma_wait_for_complete(pDev, &pDev->pStreamSgMapping[0], 5*1000);
					if (ret > 0)
					{
						dev_dbg(&(pDev->pci_devp->dev), "%s: DMA Completed (%ld)\n", __FUNCTION__, ret);
						ret= kIoctl.uiLength;
					}
					else
						pr_err("%s: %s: DMA DID NOT COMPLETE (%ld)\n", DEVICE_NAME, __FUNCTION__, ret);

					flidma_stop(pDev, &pDev->pStreamSgMapping[0]);

					// free the descriptor list....
					flidma_free_descriptors(pDev);
				}
				else
				{
					pr_err("%s: %s: failure creating descriptor list\n", DEVICE_NAME, __FUNCTION__);
					ret = -EINVAL;
				}

				FreeStreamInit(&streamInit);
			}
		}
		else
		{
			pr_err("%s: %s: failure copying ioctl from user space\n", DEVICE_NAME, __FUNCTION__);
			ret = -EINVAL;
		}
	}

	dev_dbg(&(pDev->pci_devp->dev), "%s: Leaving (%ld)\n", __FUNCTION__,ret);
	return(ret);
}

long DoStreamRead(struct file *filp, unsigned long arg)
{
	long                        ret = 0;
	FLI_CAM_IOCTL_FRAME_READ_T  kIoctl;
	fli_sg_mapping_t            *pSGMap;
	fli_cam_data_t              *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	//dev_dbg(&(pDev->pci_devp->dev), "%s: Entry\n", __FUNCTION__);
	if (arg)
	{
		if (copy_from_user((void *)(&kIoctl), (void *)(arg), sizeof(FLI_CAM_IOCTL_FRAME_READ_T)) == 0)
		{
			// Find the Descriptors to use for this transfer
			pSGMap= FindDMAMapping(pDev, kIoctl.pBuffer);
			if (pSGMap)
			{
				//pr_debug("%s: %s: Found DMA MApping 0x%px\n", DEVICE_NAME, __FUNCTION__, pSGMap);
				//pr_debug("%s: %s:    ulMappingId= 0x%lx\n", DEVICE_NAME, __FUNCTION__, pSGMap->ulMappingId);
				//pr_debug("%s: %s:    uiNumContigDescriptors= 0x%x\n", DEVICE_NAME, __FUNCTION__, pSGMap->uiNumContigDescriptors);

				// Kick off the DMA
				flidma_start(pDev, pSGMap);

				// Wait for it to complete
				ret= flidma_wait_for_complete(pDev, pSGMap, 5*1000);
				if (ret > 0)
				{
					//dev_dbg(&(pDev->pci_devp->dev), "%s: DMA Completed (%ld)\n", __FUNCTION__, ret);
					ret= kIoctl.uiLength;
				}
				else
					pr_err("%s: %s: DMA DID NOT COMPLETE (%ld)\n", DEVICE_NAME, __FUNCTION__, ret);

				flidma_stop(pDev, pSGMap);
			}
			else
			{
				pr_err("%s: %s: mapping for buffer 0x%px not found!\n", DEVICE_NAME, __FUNCTION__,kIoctl.pBuffer);
				ret = -EINVAL;
			}
		}
		else
		{
			pr_err("%s: %s: failure copying ioctl from user space\n", DEVICE_NAME, __FUNCTION__);
			ret = -EINVAL;
		}
	}

	//dev_dbg(&(pDev->pci_devp->dev), "%s: Leaving (%ld)\n", __FUNCTION__,ret);
	return(ret);
}

static long DoStreamInit(struct file *filp, unsigned long arg)
{
	long                        ret = 0;
	FLI_CAM_IOCTL_STREAM_INIT_T kStreamInit;
	fli_cam_data_t              *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);

	dev_dbg(&(pDev->pci_devp->dev), "%s: Entry\n", __FUNCTION__);

	if (arg)
	{
		if (0 == CopyStreamInitFromUser((FLI_CAM_IOCTL_STREAM_INIT_T *)arg, &kStreamInit))
		{
			// Initialize the DMA descriptors
			ret= flidma_create_descriptors(pDev, &kStreamInit);
			if (ret >= 0)
			{
				dev_dbg(&(pDev->pci_devp->dev), "%s: Streaming Initialized\n", __FUNCTION__);
				DumpDMAMappings(pDev);
			}
			else
			{
				pr_err("%s: %s: Failed to initialize the streaming interface\n", DEVICE_NAME, __FUNCTION__);
			}

			FreeStreamInit(&kStreamInit);
		}
	}

	dev_dbg(&(pDev->pci_devp->dev), "%s: Leaving (%ld)\n", __FUNCTION__,ret);

	return(ret);
}

static long DoStreamDeinit(struct file *filp)
{
	fli_cam_data_t              *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);
	dev_dbg(&(pDev->pci_devp->dev), "%s: Entry\n", __FUNCTION__);

	// Kill any DMA that might be happening- hopefully this helps us
	// recover from errors in the case of application crashes in which
	// it never did a deinit (not to mention a graceful stoppage of DMA).
	// The user API can (does) do a deinit when it opens the device.
	flidma_stop(pDev, NULL);

	// free the DMA descriptor list....
	flidma_free_descriptors(pDev);

	return(0);
}

static long DoWriteReferenceFifo(struct file *filp, unsigned long arg)
{
	long                iResult = 0;
	unsigned int        uiWordsLeft;
	unsigned int        uiWordsToCopy;
	unsigned int        *pData;
	unsigned int        *pOutBuff;
	void                *pBar;
	unsigned int        uiFifoStatus;
	FLI_CAM_IOCTL_MSG_T kIoctl;
	fli_cam_data_t *pDev;

	pDev = (fli_cam_data_t *)(filp->private_data);
	dev_dbg(&(pDev->pci_devp->dev), "%s: ENTERED\n", __FUNCTION__);
	if (arg)
	{
		if (copy_from_user((void *)(&kIoctl), (void *)(arg), sizeof(FLI_CAM_IOCTL_MSG_T)) == 0)
		{
			pBar = pDev->bar0.hdl;

			// Get the first chunk of data into kernel space
			uiWordsLeft= kIoctl.length;
			if (uiWordsLeft < FLI_IOCTL_REF_FIFO_MAX)
				uiWordsToCopy= uiWordsLeft;
			else
				uiWordsToCopy= FLI_IOCTL_REF_FIFO_MAX;
			pData= kIoctl.pData;
			iResult= copy_from_user((void *)(s_uiReferenceFifoBuffer), (void *)pData, uiWordsToCopy * 4);
			pData+= uiWordsToCopy;
			while ((iResult >= 0) && (uiWordsLeft > 0))
			{
				//pr_debug("%s: writing %d entries 0x%08x\n", __FUNCTION__,uiWordsToCopy, s_uiReferenceFifoBuffer[0]);
				// write the chunk of data to the fifo
				pOutBuff= &s_uiReferenceFifoBuffer[0];
				while (uiWordsToCopy > 0)
				{
					DoRegWrite(pDev, pBar, REG_PCIE_REFFRAME_DATA, *pOutBuff++);
					uiWordsLeft--;
					uiWordsToCopy--;
				}

				if (uiWordsLeft > 0)
				{
					// Figure out how many words we can write for the next chunk
					// We do this so we do not have to read
					// the status register for every word.  Reading is like an order of magnitude
					// slower than the writes, so this speeds things up quite a bit
					DoRegRead(pDev, pBar, REG_PCIE_REFFRAME_STATUS, &uiFifoStatus);
					uiFifoStatus &= REG_PCIE_REFFRAME_STATUS_COUNT_MASK;
					uiFifoStatus = uiFifoStatus >> 16;
					if (uiFifoStatus <= FLI_IOCTL_REF_FIFO_MAX)
					{
						uiWordsToCopy = FLI_IOCTL_REF_FIFO_MAX - uiFifoStatus;
						if (uiWordsToCopy >= uiWordsLeft)
							uiWordsToCopy = uiWordsLeft;
						if (uiWordsToCopy > 0)
						{
							iResult= copy_from_user((void *)(s_uiReferenceFifoBuffer), (void *)pData, uiWordsToCopy * 4);
							pData+= uiWordsToCopy;
						}
					}
					else
					{
						pr_err("%s: %s: FIFO over crowded with %d entries\n", DEVICE_NAME, __FUNCTION__,uiFifoStatus);
						iResult= -1;
					}
				}
			}
		}
		else
		{
			pr_err("%s: %s: copy_from_user failed\n", DEVICE_NAME, __FUNCTION__);
		}
	}
	else
	{
		pr_err("%s: %s: No argument given\n", DEVICE_NAME, __FUNCTION__);
	}

	return(iResult);
}


long FillVersionInfo(struct file *filp, DRV_INFO *pInfo)
{
   long              ret = 0;
   void *            hdl;
   fli_cam_data_t    *pDev;
   DRV_INFO          k_Info;

   pDev = (fli_cam_data_t *)(filp->private_data);
   dev_dbg(&(pDev->pci_devp->dev), "%s: Entry\n", __FUNCTION__);

   if (pInfo)
   {
		// The fpaga supports two memory regions (BAR - base address register)
		hdl = pDev->bar0.hdl;

		// Fill in the info
		memset(&k_Info,0,sizeof(DRV_INFO));
		k_Info.drvVersion= FLI_CAM_MAKE_VERSION(FLI_PCIE_MAJOR_VERSION, FLI_PCIE_MINOR_VERSION, FLI_PCIE_BUILD_NUMBER);
		ret= DoRegRead(pDev, pDev->bar0.hdl, 0, &k_Info.hwFibreId);
		if (0 == ret)
			ret= DoRegRead(pDev, pDev->bar0.hdl, 4, &k_Info.hwFibreVersion);
		if (0 == ret)
			ret= DoRegRead(pDev, pDev->bar0.hdl, 0x1000, &k_Info.hwPcieId);
		if (0 == ret)
			ret= DoRegRead(pDev, pDev->bar0.hdl, 0x1004, &k_Info.hwPcieVersion);

		if (0 == ret)
		{
			if (copy_to_user((void *)(pInfo), (void *)(&k_Info), sizeof(DRV_INFO)) != 0)
			{
				pr_err("%s: %s: copy_to_user error\n", DEVICE_NAME, __FUNCTION__);
				ret = -EFAULT;
			}
		}
   }
   else
   {
	  pr_err("%s: %s:  NULL info pointer\n", DEVICE_NAME, __FUNCTION__);
	  ret = -EINVAL;
   }

   return ret;
}

fli_sg_mapping_t *FindDMAMapping(fli_cam_data_t *pDev, unsigned char *pBuffer)
{
	fli_sg_mapping_t *pMap;
	int i;

	if ((NULL == pDev) || (NULL == pBuffer))
		return(NULL);

	pMap= NULL;
	for (i=0; (i<pDev->uiNumStreamMappings) && (NULL == pMap); ++i)
	{
		if (pDev->pStreamSgMapping[i].ulMappingId == (unsigned long)pBuffer)
			pMap= &pDev->pStreamSgMapping[i];
	}

	return(pMap);
}

void DumpDMAMappings(fli_cam_data_t *pDev)
{
	int i;

	if (NULL == pDev)
		return;

	pr_debug("%s: %s: Num DMA Mappings= %d\n", DEVICE_NAME, __FUNCTION__, pDev->uiNumStreamMappings);
	for (i=0; i<pDev->uiNumStreamMappings; ++i)
	{
		pr_debug("%s: %s: DMA Mapping 0x%px\n", DEVICE_NAME, __FUNCTION__, &pDev->pStreamSgMapping[i]);
		pr_debug("%s: %s:    ulMappingId= 0x%lx\n", DEVICE_NAME, __FUNCTION__, pDev->pStreamSgMapping[i].ulMappingId);
		pr_debug("%s: %s:    uiNumContigDescriptors= 0x%x\n", DEVICE_NAME, __FUNCTION__, pDev->pStreamSgMapping[i].uiNumContigDescriptors);
	}
}


void DumpConfig(struct pci_dev *pDev)
{
	unsigned char cVal;
	unsigned short usVal;
	unsigned int   uiVal;

	pci_read_config_word(pDev,PCI_VENDOR_ID,&usVal);
	dev_dbg(&(pDev->dev), "PCI_VENDOR_ID 0x%04X\n", usVal);
	pci_read_config_word(pDev,PCI_DEVICE_ID,&usVal);
	dev_dbg(&(pDev->dev), "PCI_DEVICE_ID 0x%04X\n", usVal);
	pci_read_config_word(pDev,PCI_STATUS,&usVal);
	dev_dbg(&(pDev->dev), "PCI_STATUS 0x%04X\n", usVal);
	pci_read_config_dword(pDev,PCI_CLASS_REVISION,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_CLASS_REVISION 0x%08X\n", uiVal);

	pci_read_config_byte(pDev,PCI_CACHE_LINE_SIZE,&cVal);
	dev_dbg(&(pDev->dev), "PCI_CACHE_LINE_SIZE 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_LATENCY_TIMER,&cVal);
	dev_dbg(&(pDev->dev), "PCI_LATENCY_TIMER 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_HEADER_TYPE,&cVal);
	dev_dbg(&(pDev->dev), "PCI_HEADER_TYPE 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_BIST,&cVal);
	dev_dbg(&(pDev->dev), "PCI_BIST 0x%02X\n", cVal);

	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_0,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_0 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_1,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_1 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_2,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_2 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_3,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_3 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_4,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_4 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_BASE_ADDRESS_5,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_BASE_ADDRESS_5 0x%08X\n", uiVal);

	pci_read_config_word(pDev,PCI_CARDBUS_CIS,&usVal);
	dev_dbg(&(pDev->dev), "PCI_CARDBUS_CIS 0x%04X\n", usVal);
	pci_read_config_word(pDev,PCI_SUBSYSTEM_VENDOR_ID,&usVal);
	dev_dbg(&(pDev->dev), "PCI_SUBSYSTEM_VENDOR_ID 0x%04X\n", usVal);
	pci_read_config_word(pDev,PCI_SUBSYSTEM_ID,&usVal);
	dev_dbg(&(pDev->dev), "PCI_SUBSYSTEM_ID 0x%04X\n", usVal);
	pci_read_config_dword(pDev,PCI_ROM_ADDRESS,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_ROM_ADDRESS 0x%08X\n", uiVal);
	pci_read_config_dword(pDev,PCI_CAPABILITY_LIST,&uiVal);
	dev_dbg(&(pDev->dev), "PCI_CAPABILITY_LIST 0x%08X\n", uiVal);

	pci_read_config_byte(pDev,PCI_INTERRUPT_LINE,&cVal);
	dev_dbg(&(pDev->dev), "PCI_INTERRUPT_LINE 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_INTERRUPT_PIN,&cVal);
	dev_dbg(&(pDev->dev), "PCI_INTERRUPT_PIN 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_MIN_GNT,&cVal);
	dev_dbg(&(pDev->dev), "PCI_MIN_GNT 0x%02X\n", cVal);
	pci_read_config_byte(pDev,PCI_MAX_LAT,&cVal);
	dev_dbg(&(pDev->dev), "PCI_MAX_LAT 0x%02X\n", cVal);
}

//module_init(fli_cam_init);
//module_exit(fli_cam_exit);
//MODULE_LICENSE("GPL");

