/*
 * flipcie_dma.c
 *
 *  Created on: Aug 21, 2020
 *      Author: tom
 */
#include <asm/byteorder.h>
#include <asm/cacheflush.h>

#include "dma_engine.h"
#include "flipcie_dma.h"
#include "flipcie_regs.h"

extern int doCreateSgDescriptorList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap, unsigned char *pBuffer, unsigned long uiLength);
extern int doFreeSgDescriptorList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap);
static void assign_pages_to_sgl(unsigned char *pPinnedUserBuffer, unsigned long uiUserLength,  int iNumPages, struct page **ppPages, struct scatterlist *sg);
static void free_sg_mapping(fli_cam_data_t *pDevs, fli_sg_mapping_t *pSGMap);
static void ResetSGTransaction(fli_sg_mapping_t *pSGTrans);
static void DumpSGList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGTrans);
static uint64_t ktimeDiffMS(ktime_t kStart, ktime_t kEnd);

int flidma_probe(struct pci_dev *pci_devp, fli_cam_data_t *pCamDev)
{
	int iResult;

	iResult= -1;

	pCamDev->pStreamSgMapping= NULL;
	pCamDev->uiNumStreamMappings= 0;

	pCamDev->bar1.start = pci_resource_start(pci_devp, BAR_1 );
	if (pCamDev->bar1.start != 0)
	{
	    dev_dbg(&(pci_devp->dev), "%s: Using BAR 1\n", __FUNCTION__);

		// Bar 1 appears OK- use it
	    pCamDev->bar1.len   = pci_resource_len(pci_devp, BAR_1 );
	    pCamDev->bar1.flags = pci_resource_flags(pci_devp, BAR_1 );

	    pCamDev->bar1.hdl = pci_iomap(pci_devp, BAR_1, 0);

		iResult= 0;
	}
	else
	{
		// We are not sure why- but BAR 1 is not being enumerated correctly and it appears
		// to be being put in BAR 2 when this happens, so we will use BAR 2 instead
		pCamDev->bar1.start = pci_resource_start(pci_devp, BAR_2 );
	    dev_dbg(&(pci_devp->dev), "%s: BAR 2 resource start 0x%lX\n", __FUNCTION__ ,pCamDev->bar1.start);
		if (pCamDev->bar1.start != 0)
		{
		    dev_dbg(&(pci_devp->dev), "%s: Using BAR 2\n", __FUNCTION__);

			// Bar 2 appears OK- use it
		    pCamDev->bar1.len   = pci_resource_len(pci_devp, BAR_2 );
		    pCamDev->bar1.flags = pci_resource_flags(pci_devp, BAR_2 );

		    pCamDev->bar1.hdl = pci_iomap(pci_devp, BAR_2, 0);

			iResult= 0;
		}
		else
		{
			pCamDev->bar1.len   = pci_resource_len(pci_devp, BAR_2 );
			pCamDev->bar1.flags = pci_resource_flags(pci_devp, BAR_2 );

			pCamDev->bar1.hdl = pci_iomap(pci_devp, BAR_2, 0);
			pr_err("Error BAR 2 resource start  returned 0! (%s)\n", DEVICE_NAME);
			pr_err("Device info DMA Registers: start 0x%lX, len 0x%lX, flags 0x%lX , bar1.hdl 0x%LX (%s)\n",
					pCamDev->bar1.start, pCamDev->bar1.len, pCamDev->bar1.flags, (uint64_t)pCamDev->bar1.hdl, DEVICE_NAME);

		}
	}

	if (iResult >= 0)
		iResult= dma_engine_probe(pci_devp,pCamDev);

    dev_dbg(&(pci_devp->dev), "%s: Device info DMA Registers: start 0x%lX, len 0x%lX, flags 0x%lX , bar1.hdl 0x%LX\n", __FUNCTION__,
    		pCamDev->bar1.start, pCamDev->bar1.len, pCamDev->bar1.flags, (uint64_t)pCamDev->bar1.hdl);

    return(iResult);
}

int flidma_remove(struct pci_dev *pPciDev, fli_cam_data_t *pCamDev)
{
	int iResult;

	iResult= dma_engine_remove(pPciDev,pCamDev);
	return(iResult);
}

int flidma_create_descriptors(fli_cam_data_t *pDev, FLI_CAM_IOCTL_STREAM_INIT_T *pStreamInit)
{
	int iResult;
	int i;

	// allocate the mapping structures- note we protect against multiple threads
	// trying to allocate while another thread already has something going
	// We do not want to delete DMA buffers while streaming for example- that would be bad...
	iResult= mutex_lock_interruptible(&pDev->mutexDMAMapping);
	if (0 == iResult)
	{
		if (NULL == pDev->pStreamSgMapping)
		{
			pDev->pStreamSgMapping= (fli_sg_mapping_t *)kcalloc(pStreamInit->uiNumBuffers, sizeof(fli_sg_mapping_t), GFP_KERNEL);
			mutex_unlock(&pDev->mutexDMAMapping);

			if (pDev->pStreamSgMapping)
			{
				pDev->uiNumStreamMappings= pStreamInit->uiNumBuffers;
				for (i=0; i<pStreamInit->uiNumBuffers; ++i)
				{
					dev_dbg(&(pDev->pci_devp->dev), "%s: Creating descriptors for buffer %d\n", __FUNCTION__, i);
					iResult= doCreateSgDescriptorList(pDev, &pDev->pStreamSgMapping[i], pStreamInit->ppBuffers[i], pStreamInit->uiLength) ;
					if (iResult >= 0)
					{
						// Now create the DMA scatter-gather for the DMA engine on the hardware
						iResult= dma_build_contiguous_sgdescriptor_list(pDev, &pDev->pStreamSgMapping[i]);
						if (iResult < 0)
						{
							pr_err("%s: Error creating HW Descriptor List\n", __FUNCTION__);
							flidma_free_descriptors(pDev);
						}
					}
					else
					{
						pr_err("%s: Error creating SG List\n", __FUNCTION__);
					}
					dev_dbg(&(pDev->pci_devp->dev), "%s: Completed descriptors for buffer %d\n", __FUNCTION__, i);
				}
			}
			else
			{
				pr_err("%s: Error Allocating Mapping list for %d buffers\n", __FUNCTION__,pStreamInit->uiNumBuffers);
				pDev->uiNumStreamMappings= 0;
				iResult= -1;
			}
		}
		else
		{
			mutex_unlock(&pDev->mutexDMAMapping);
			pr_err("%s: DMA Mappings already in use- deinit the streamer perhaps.\n", __FUNCTION__);
			iResult= -1;
		}
	}
	else
	{
		pr_err("%s: Error %d obtaining DMA MApping lock- must have been cancelled.\n", __FUNCTION__, iResult);
	}

	return(iResult);
}

int flidma_free_descriptors(fli_cam_data_t *pDev)
{
	int iResult;
	int i;

	iResult= mutex_lock_interruptible(&pDev->mutexDMAMapping);
	if (0 == iResult)
	{
		if (pDev->pStreamSgMapping)
		{
			for (i=0; i< pDev->uiNumStreamMappings; ++i)
			{
				// free the contiguous list
				dma_free_contiguous_sgdescriptor_list(pDev, &pDev->pStreamSgMapping[i]);

				// free the sg descriptor list....
				doFreeSgDescriptorList(pDev, &pDev->pStreamSgMapping[i]);
			}

			kfree(pDev->pStreamSgMapping);
			pDev->pStreamSgMapping= NULL;
		}
		pDev->uiNumStreamMappings= 0;

		mutex_unlock(&pDev->mutexDMAMapping);
	}
	else
	{
		pr_err("%s: Error %d obtaining DMA MApping lock- must have been cancelled.\n", __FUNCTION__, iResult);
	}

	return(iResult);
}



int flidma_start(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	int iResult;

	if (NULL == pDev)
		return(-1);

	iResult= -1;

	iResult= dma_engine_start_xfer(pDev, pSGMap);

	return(iResult);
}

int flidma_stop(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	int iResult;

	if (NULL == pDev)
		return(-1);

	iResult= -1;

	iResult= dma_engine_stop_xfer(pDev, pSGMap);

	return(iResult);
}




int flidma_wait_for_complete(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap, uint64_t uiTimeoutMS)
{
	ktime_t ktimeStart;
	int     bDone;

	ktimeStart = ktime_get();
	bDone= 0;
	do
	{
		bDone= dma_engine_is_xfer_complete(pDev, pSGMap);
	} while ((0 == bDone) && (ktimeDiffMS(ktimeStart, ktime_get()) < uiTimeoutMS));

	//dev_dbg(&(pDev->pci_devp->dev), "%s: Dumping dma regs...\n", __FUNCTION__);
	if (!(bDone))
		dma_engine_dump_regs(pDev);

	return(bDone);
}


//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
/// Static function Implementations
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
int doCreateSgDescriptorList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap, unsigned char *pBuffer, unsigned long uiLength)
{
	long       ret = 0;
	int pages_calc;
	struct sg_table *pSGTable;

	//dev_dbg(&(pDev->pci_devp->dev), "Entry %s (%s)\n", __FUNCTION__, DEVICE_NAME);

	// Get the convenience pointers set up to the transaction table
	pSGTable = &(pSGMap->sgTable);

	// Calc # of pages in the buffer
	pages_calc = (((unsigned long)pBuffer + uiLength + PAGE_SIZE -1) - ((unsigned long)pBuffer & PAGE_MASK)) >> PAGE_SHIFT;

	// Create the empty scatter gather table
	if (sg_alloc_table(pSGTable, pages_calc, GFP_KERNEL) == 0)
	{
		pSGMap->bSGAllocated = true;

		// Allocate an array of pages
		pSGMap->pages = kcalloc(pages_calc, sizeof(struct page *), GFP_KERNEL);
		if (pSGMap->pages)
		{
			// pin user pages in memory
			pSGMap->pages_got = get_user_pages_fast((unsigned long)pBuffer, pages_calc, 1/* write */, pSGMap->pages);
			if (pSGMap->pages_got > 0)
			{
				// get_user_pages_fast() shouldn't return fewer than the requested number of pages,
				// Since later code assumes the size of the pages, need to know we got what we expect
				if (pSGMap->pages_got == pages_calc)
				{
					// assign the pages to an entry in the scatter-gather list
					assign_pages_to_sgl(pBuffer,uiLength, pSGMap->pages_got, pSGMap->pages, pSGTable->sgl);

					// Finalize the collection of pages, user buffer chunks, and scatter-gather table entries into
					// a sg table, this function may combine adjacent pages, so save the number of entries returned
					// I think this could also be a call to dma_map_sg, pci_map_sg may be on it's way out.....
					//pSGMap->numEntries = pci_map_sg(pDev->pci_devp, pSGTable->sgl, pSGTable->orig_nents, DMA_FROM_DEVICE);
					pSGMap->numEntries = dma_map_sg(&pDev->pci_devp->dev, pSGTable->sgl, pSGTable->orig_nents, DMA_FROM_DEVICE);
					if (pSGMap->numEntries > 0)
					{
						pSGMap->ulMappingId= (unsigned long)pBuffer;
						DumpSGList(pDev, pSGMap);
					}
					else
					{
						pr_err("pci_map_sg error (%s)\n", DEVICE_NAME);
						ret = -EIO;
					}
				}
				else
				{
				   pr_err("get_user_pages_fast wrong number of pages got 0x%x 0x%x (%s)\n", pages_calc, pSGMap->pages_got, DEVICE_NAME);
				   ret = -EFAULT;
				}
			}
			else
			{
				pr_err("get_user_pages_fast error (%s)\n", DEVICE_NAME);
				ret = -EFAULT;
			}
		}
		else
		{
			pr_err("kcalloc error, pages (%s)\n", DEVICE_NAME);
			ret = -ENOMEM;
		}
	}
	else
	{
		pr_err("sg_alloc_table error (%s)\n", DEVICE_NAME);
		ret = -ENOMEM;
	}

	//dev_dbg(&(pDev->pci_devp->dev), "Exit %s (%s)\n", __FUNCTION__, DEVICE_NAME);

	return(ret);
}


int doFreeSgDescriptorList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	free_sg_mapping(pDev, pSGMap);
	return(0);
}

void assign_pages_to_sgl(unsigned char *pPinnedUserBuffer, unsigned long uiUserLength, int iNumPages, struct page **ppPages, struct scatterlist *sg)
{
   unsigned long len_remain;
   unsigned int offset;
   unsigned int nbytes;
   int i;

   //dev_dbg(&(this_devp->pci_devp->dev), "Entry %s (%s)\n", __FUNCTION__, DEVICE_NAME);

   len_remain = uiUserLength;
   for (i = 0; i < iNumPages; i++, sg = sg_next(sg))
   {
      offset = offset_in_page(pPinnedUserBuffer);
      nbytes = min_t(unsigned int, PAGE_SIZE - offset, len_remain);

      flush_dcache_page(ppPages[i]);
      sg_set_page(sg, ppPages[i], nbytes, offset);

      pPinnedUserBuffer += nbytes;
      len_remain -= nbytes;
   }

   // After assigning all the length to buffers, none should be left
   if (len_remain)
   {
      pr_err("%s: Data leftover %lu (0x%lx) of %lu (0x%lx) after setting sg pages (%s)\n",__FUNCTION__, len_remain, len_remain, uiUserLength, uiUserLength, DEVICE_NAME);
   }

   //dev_dbg(&(this_devp->pci_devp->dev), "Exit %s (%s)\n", __FUNCTION__, DEVICE_NAME);
}

void free_sg_mapping(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
   int              idx;
   struct sg_table *pSGtable;

   //dev_dbg(&(pDev->pci_devp->dev), "Entry %s (%s)\n", __FUNCTION__, DEVICE_NAME);
   pSGtable = &(pSGMap->sgTable);
   if (pSGMap->numEntries)
   {
	      //pci_unmap_sg(pDev->pci_devp, pSGtable->sgl, pSGtable->orig_nents, DMA_FROM_DEVICE);
	      dma_unmap_sg(&pDev->pci_devp->dev, pSGtable->sgl, pSGtable->orig_nents, DMA_FROM_DEVICE);
   }

   if (pSGMap->pages_got)
   {
      if (pSGMap->pages)
      {
         // un-pin the pages
         for (idx = 0; idx < pSGMap->pages_got; idx++)
         {
            put_page(pSGMap->pages[idx]);
         }
      }
      else
      {
         pr_err("dma pages got 0x%X, without pages pointer (%s)\n", pSGMap->pages_got, DEVICE_NAME);
      }
   }

   if (pSGMap->pages)
   {
      kfree(pSGMap->pages);
      pSGMap->pages= NULL;
   }

   if (pSGMap->bSGAllocated)
   {
      sg_free_table(&(pSGMap->sgTable));
      pSGMap->bSGAllocated= false;
   }

   ResetSGTransaction(pSGMap);

}


void ResetSGTransaction(fli_sg_mapping_t *pSGMap)
{
	pSGMap->numEntries= 0;
	pSGMap->pages_got= 0;
	pSGMap->pages= NULL;
	pSGMap->bSGAllocated= false;
}

void DumpSGList(fli_cam_data_t *pDev, fli_sg_mapping_t *pSGMap)
{
	struct scatterlist *pSG;
	int i;
	uint64_t uiLastDmaAddress;
	uint64_t uiLastDmaLength;
	uint64_t uiTotalDmaLength;

	if (NULL == pSGMap)
	{
		dev_dbg(&(pDev->pci_devp->dev), "%s: NULL SG List\n", __FUNCTION__);
	}
	dev_dbg(&(pDev->pci_devp->dev), "%s: SG List\n", __FUNCTION__);
	dev_dbg(&(pDev->pci_devp->dev), "%s:     pages_got= %d, numEntries= %d\n", __FUNCTION__,pSGMap->pages_got, pSGMap->numEntries);
	dev_dbg(&(pDev->pci_devp->dev), "%s:     sgTable.nents= %d, sgTable.orig_nents= %d\n", __FUNCTION__,pSGMap->sgTable.nents, pSGMap->sgTable.orig_nents);
	pSG= pSGMap->sgTable.sgl;
	uiTotalDmaLength= 0;
	uiLastDmaAddress= 0;
	uiTotalDmaLength= 0;
	for (i=0; i<pSGMap->sgTable.nents; ++i, pSG= sg_next(pSG))
	{
		if (0 == i)
		{
			dev_dbg(&(pDev->pci_devp->dev), "%s: %d. page_link= 0x%lx, offset= 0x%x, length= 0x%x, dma_address= 0x%llx\n",
											 __FUNCTION__,i,pSG->page_link, pSG->offset,sg_dma_len(pSG),sg_dma_address(pSG));
			uiLastDmaAddress= sg_dma_address(pSG);
			uiLastDmaLength= sg_dma_len(pSG);
			uiTotalDmaLength+= uiLastDmaLength;
		}
		else
		{
			if (sg_dma_address(pSG) != (uiLastDmaAddress + uiLastDmaLength))
			{
				dev_dbg(&(pDev->pci_devp->dev), "%s: %d. not contiguous with next: length= 0x%llx, dma_address= 0x%llx\n",
												 __FUNCTION__,i-1,uiLastDmaLength,uiLastDmaAddress);
				dev_dbg(&(pDev->pci_devp->dev), "%s: %d. page_link= 0x%lx, offset= 0x%x, length= 0x%x, dma_address= 0x%llx\n",
												 __FUNCTION__,i,pSG->page_link, pSG->offset,sg_dma_len(pSG),sg_dma_address(pSG));
			}
			else
			{
				if (i >= (pSGMap->sgTable.nents - 2))
				{
					dev_dbg(&(pDev->pci_devp->dev), "%s: %d. page_link= 0x%lx, offset= 0x%x, length= 0x%x, dma_address= 0x%llx\n",
													 __FUNCTION__,i,pSG->page_link, pSG->offset,sg_dma_len(pSG),sg_dma_address(pSG));
				}
			}

			uiLastDmaAddress= sg_dma_address(pSG);
			uiLastDmaLength= sg_dma_len(pSG);
			uiTotalDmaLength+= uiLastDmaLength;
		}
	}
	dev_dbg(&(pDev->pci_devp->dev), "%s: Total DMA length= 0x%llx\n", __FUNCTION__,uiTotalDmaLength);
}


uint64_t ktimeDiffMS(ktime_t kStart, ktime_t kEnd)
{
	uint64_t uiDiff;

	if (kEnd >= kStart)
		uiDiff= kEnd - kStart;
	else
		uiDiff= (uint64_t)0xFFFFFFFFFFFFFFFF - kStart + kEnd;

	uiDiff= uiDiff / 1000000;
	return(uiDiff);
}

