//***************************************************************
/// @file flipcie_ioctl.h
/// @brief User side interface for the FLI Host Card Device Driver
/// @author      Tatakis
/// @date        
/// @copyright  
///
//
//
//***************************************************************

#ifndef FLIPCIE_IOCTL_H_
#define FLIPCIE_IOCTL_H_
#include "asm/ioctl.h"
#include <linux/types.h>

/// Base name of device in /dev directory
#define FLI_DEVICE_NAME "flipcie_cam"
#define FLI_DEVICE_PATH "/dev/" FLI_DEVICE_NAME

/// Base name of device in /dev directory
#define DEVICE_NAME FLI_DEVICE_NAME

#define BIT_N(n)  (1 << (n))

/// FPGA Register Read or Write
typedef struct
{
   unsigned int   reg;  ///< address of the register to write
   unsigned int   val;  ///< data to read from, or write to, the register
   unsigned int   bar;  ///< 0,1, .. the PCI Base Address Register on the device
} FLI_CAM_IOCTL_RW_REG_T;

/// Send/Recv Message
#define FLI_IOCTL_MESSAGE_LENGTH_MAX (512)
#define FLI_IOCTL_MESSAGE_MIN_DWORDS (4)
typedef struct
{
	unsigned int   length;  ///< Length of data- number of 32 bit words
	unsigned int   *pData;  ///< data to send/buffer to receive
} FLI_CAM_IOCTL_MSG_T;

typedef struct
{
	unsigned int  uiLength;  // Length in bytes
	unsigned char *pBuffer;  // User buffer to read into
} FLI_CAM_IOCTL_FRAME_READ_T;

typedef struct
{
	unsigned int  uiLength;      // Length of buffers in bytes
	unsigned int  uiNumBuffers;  // Number of pBuffer pointers
	unsigned char **ppBuffers;   // User buffers to read into
} FLI_CAM_IOCTL_STREAM_INIT_T;

//------------------------------------------------------------------
/**
 * Device Driver Information.
 * Contains the version and operating mode for the FLI PCIe device driver and hardware
 *
 * Unless specified otherwise, all version number are 4 byte values: 0xMMmmBBPP,
 * where MM=Major version, mm=Minor version, BB=Build number, PP=Patch number.
 *
 * @sa FLI_REG_IOCTL_VERSION
 * @todo Define hardware status flags
 */
typedef struct
{
	unsigned int        drvVersion;     ///< Driver version, MMmmBBPP
	unsigned int        hwFibreId;      ///< Fibre Side FPGA Device ID
	unsigned int        hwFibreVersion; ///< Fibre Side FPGA Version
	unsigned int        hwPcieId;       ///< PCIE Side FPGA Device ID
	unsigned int        hwPcieVersion;  ///< PCIE Side FPGA Version
	unsigned int        hwPciBus;       ///< I/F board PCIe bus number
	unsigned int        hwPciDevice;    ///< I/F board PCIe device number
	unsigned int        hwPciFunction;  ///< I/f board PCIe function number
} DRV_INFO;

/// @name IOCTL Number Base
/// Some ioctl numbers in the 'u' range are theoretically available
/// based on Documentation/ioctl/ioctl-number.txt
///
/// Start the I numbers after the range 00-1F
/// @{
#define FLI_CAM_IOCTL_NUM_TYPE ('u')  ///< IOCTL Range base
#define FLI_CAM_IOCTL_NUM_BASE (0x41) ///< IOCTL Range offset
///@}


/// @name IOCTL Commands
/// @sa flipcie_reg_ioctl()
/// @{
#define FLI_CAM_IOCTL_READ_REG     _IOR(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 0), FLI_CAM_IOCTL_RW_REG_T *) ///< Read a register
#define FLI_CAM_IOCTL_WRITE_REG    _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 1), FLI_CAM_IOCTL_RW_REG_T *) ///< Write a register
#define FLI_CAM_IOCTL_VERSION      _IO(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 2)) ///< The ioctl call will return a 32 bit value containing the driver version

#define FLI_CAM_IOCTL_SEND_MESSAGE    _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 3), FLI_CAM_IOCTL_MSG_T *) ///< Send/Recv Message
#define FLI_CAM_IOCTL_GET_RESPONSE    _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 4), FLI_CAM_IOCTL_MSG_T *) ///< Send/Recv Message

#define FLI_CAM_IOCTL_FRAME_READ    _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 5), FLI_CAM_IOCTL_FRAME_READ_T *) ///< Read a single frame
#define FLI_CAM_IOCTL_STREAM_INIT   _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 6), FLI_CAM_IOCTL_STREAM_INIT_T *) ///< Initialize streaming
#define FLI_CAM_IOCTL_STREAM_DEINIT _IO(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 7)) ///< Deinitialize streaming
#define FLI_CAM_IOCTL_STREAM_READ    _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 8), FLI_CAM_IOCTL_FRAME_READ_T *) ///< Read a streaming frame

#define FLI_CAM_IOCTL_WRITE_REF_FIFO _IOW(FLI_CAM_IOCTL_NUM_TYPE, (FLI_CAM_IOCTL_NUM_BASE + 9), FLI_CAM_IOCTL_MSG_T *) ///< Write the reference frame fifo

///@}

/// @name Driver Version
/// The driver version will consist of three fields
/// - bits 31:24 - major version
/// - bits 23:16 - minor version
/// - bits 15:8  - build number
/// - bits 7:0   - reserved
/// @{
#define FLI_CAM_MAKE_VERSION(major,minor,build) ( ((major & 0xff) << 16) \
                                                   + ((minor & 0xff) << 8 ) \
                                                   + ((build & 0xff)      ))

#define FLI_CAM_GET_MAJOR_VERSION(version) ((version & 0x00ff0000) >> 16)
#define FLI_CAM_GET_MINOR_VERSION(version) ((version & 0x0000ff00) >> 8 )
#define FLI_CAM_GET_BUILD_NUMBER(version)  ((version & 0x000000ff)      )
///@}


#endif /* FLIPCIE_IOCTL_H_ */
