//***************************************************************
/// @file fli_main.c
/// @brief register read/write handling for the FLI Host card device driver
/// @author      ttatakis
/// @date
/// @copyright
///
//
//
//***************************************************************

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/version.h>
//#include <linux/delay.h>
//#include <linux/kthread.h>

#include <asm/byteorder.h>
#include <asm/cacheflush.h>

#include "flipcie_priv.h"
#include "flipcie_regs.h"
#include "flipcie_ioctl.h"

#define VENDOR_ID_XILINX (0x10EE)
#define DEVICE_ID_HOST_BOARD (0x4631)


// Each image and dma device will have a the same major number
// and a different minor number
// Device         Minor #
// fli_cam0      0
// fli_image0    1
// fli_cam1      2
// fli_image1    3
#define NUM_FLI_DEVICE_TOTAL (NUM_FLI_PCIE_DEVICE*2)
#define BASE_MINOR_NUM      (0)

static struct pci_device_id ids[] = {
   { PCI_DEVICE(VENDOR_ID_XILINX, DEVICE_ID_HOST_BOARD), },
   { 0, }
};
MODULE_DEVICE_TABLE(pci, ids);

/// Device instance data
typedef struct
{
   int      idx;            ///< instance # of device data
   int      major;          ///< device major number
   int      minor;          ///< device minor number
   bool     available;      ///< indicates if device data is available
} fli_pcie_dev_data_t;

static int my_dev_uevent(const struct device *dev, struct kobj_uevent_env *env);

/// Contains device major number
static dev_t fli_pcie_dev_num;
struct class *fli_class;

fli_pcie_dev_data_t * pFli_dev_data[NUM_FLI_PCIE_DEVICE];

// Called by linux when a device with matching numbers is discovered
static int probe(struct pci_dev *pci_devp, const struct pci_device_id *id)
{
   int             ret;
   int             idx;
   fli_pcie_dev_data_t * host_datap;

   dev_dbg(&(pci_devp->dev), "%s: probe ENTERED\n", DEVICE_NAME);

   if(pci_enable_device(pci_devp))
   {
      dev_err(&pci_devp->dev, "%s: can't enable PCI device\n", DEVICE_NAME);
      return -ENODEV;
   }

   // Enable bus mastering for DMA
   pci_set_master(pci_devp);

   // Take ownership of the device memory regions
#if 0
   ret = pci_request_regions(pci_devp, DEVICE_NAME);
   if (ret)
   {
      pci_disable_device(pci_devp);
      pr_err("Error pci_request_regions %d (%s)\n", ret, DEVICE_NAME);
      return ret;
   }
#else
   ret = pci_request_region(pci_devp, 0, DEVICE_NAME);
   if (ret)
   {
      pr_err("%s: Error pci_request_region 0 failed %d\n", DEVICE_NAME, ret);
   }
   //ret = pci_request_region(pci_devp, 1, DEVICE_NAME);
   //if (ret)
   //{
   //   pr_err("Error pci_request_region 1 failed %d (%s)\n", ret, DEVICE_NAME);
  // }
   ret = pci_request_region(pci_devp, 2, DEVICE_NAME);
   if (ret)
   {
      pr_err("%s: Error pci_request_region 2 failed %d\n", DEVICE_NAME, ret);
   }
#endif

    // Find an available set of device data
    host_datap = NULL;
    for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
    {
      // if device data is available
      if (pFli_dev_data[idx]->available)
      {
         host_datap = pFli_dev_data[idx];
         host_datap->available = false;
         break;
      }
    }

    // If we found available device data
    if (host_datap)
    {
      // Save ptr to our data, for use in release
      dev_set_drvdata(&(pci_devp->dev), (void *)host_datap);

      // Initialize major and minor numbers to use as devices are found
      host_datap->major = MAJOR(fli_pcie_dev_num);

      // Save/assign minor number
      host_datap->idx = idx;
      host_datap->minor = idx;
#if 1
      ret = fli_cam_probe(pci_devp, fli_class, idx, host_datap->major, host_datap->minor);
      if (ret == 0)
      {
         //ret = fli_image_probe(pci_devp, fli_class, idx, host_datap->major, host_datap->image_minor);
         if (ret)
         {
            fli_cam_remove(pci_devp, fli_class, idx);
            pr_err("%s: Fail fli_image_probe\n", DEVICE_NAME);
         }
      }
      else
      {
         pr_err("%s: Fail fli_cam_probe\n", DEVICE_NAME);
      }
#endif

   }
   else
   {
      pr_err("%s: Error no device data avail\n", DEVICE_NAME);
      ret = -ENOMEM;
   }

   if (ret)
   {
      pr_err("%s: Probe fail %d %d %d\n", DEVICE_NAME, idx, host_datap->major, host_datap->minor);
      unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
      pci_release_regions(pci_devp);
      pci_disable_device(pci_devp);
   }

   return 0;
} // probe

static void remove(struct pci_dev *pci_devp)
{
   fli_pcie_dev_data_t * host_datap;

   dev_dbg(&(pci_devp->dev), "remove (%s)\n", DEVICE_NAME);

   host_datap = (fli_pcie_dev_data_t *)dev_get_drvdata(&(pci_devp->dev));

   //dev_dbg(&(pci_devp->dev), "device number, major 0x%X, minor 0x%X (%s)\n", this_devp->major, this_devp->minor, DEVICE_NAME);

   fli_cam_remove(pci_devp, fli_class, host_datap->idx);

   host_datap->available = true; // Mark data as available

   // clean up any allocated resources
   pci_release_regions(pci_devp);
   pci_disable_device(pci_devp);

} // remove

static struct pci_driver pci_driver = {
   .name = DEVICE_NAME,
   .id_table = ids,
   .probe = probe,
   .remove = remove,
};

static int __init fli_pcie_init(void)
{
   int   ret;
   int   idx;

   pr_debug("%s: Loading\n", DEVICE_NAME);

   // Allocate device major number
   ret = alloc_chrdev_region( &fli_pcie_dev_num, BASE_MINOR_NUM, NUM_FLI_DEVICE_TOTAL, DEVICE_NAME);
   if (ret < 0)
   {
      pr_err("Fail alloc_chrdev_region (%s)\n", DEVICE_NAME);
      return (ret);
   }

   pr_debug("fli_pcie_dev_num  major 0x%X, minor 0x%X (%s)\n", MAJOR(fli_pcie_dev_num), MINOR(fli_pcie_dev_num), DEVICE_NAME);

   for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
   {
      // allocate device data
      pFli_dev_data[idx] = kmalloc(sizeof(fli_pcie_dev_data_t), GFP_KERNEL);
      if (!pFli_dev_data[idx])
      {
         pr_err("Fail kmalloc device data %d  (%s)\n", idx, DEVICE_NAME);

         // Free any allocated device data memory, starting at the previous entry
         while (idx)
         {
            kfree(pFli_dev_data[--idx]);
         }
         unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
         return (-ENOMEM);
      }

      // Mark device data as available
      pFli_dev_data[idx]->available = true;
   }

   ret = fli_cam_init();
   if (ret < 0)
   {
      pr_err("Fail fli_cam_init (%s)\n", DEVICE_NAME);
      unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
      return (ret);
   }

#if 0
   ret = fli_image_init();
   if (ret < 0)
   {
      pr_err("Fail fli_image_init (%s)\n", DEVICE_NAME);
      unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
      return (ret);
   }
#endif

   // Make sysfs entries

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0))
   fli_class = class_create( THIS_MODULE, DEVICE_NAME);
#else
   fli_class = class_create(DEVICE_NAME);
#endif
    if ( IS_ERR(fli_class))
   {
      pr_err("Error class_create %ld (%s)\n", (long)fli_class, DEVICE_NAME);
      unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
      return (-ENODEV);
   }
   fli_class->dev_uevent= my_dev_uevent;
   //pr_debug("fli_pcie(0x%X,0x%X): fli_class->dev_uevent:  %ld\n", MAJOR(fli_pcie_dev_num), MINOR(fli_pcie_dev_num), (long)fli_class->dev_uevent);


   ret = pci_register_driver(&pci_driver);
   if (ret < 0)
   {
      pr_err("Error pci_register_driver %d (%s)\n", ret, DEVICE_NAME);
      unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
      class_destroy(fli_class);
      return (ret);
   }

   pr_debug("%s: Loaded\n", DEVICE_NAME);
   return(0);
} // fli_pcie_init

static int my_dev_uevent(const struct device *dev, struct kobj_uevent_env *env)
{
    add_uevent_var(env, "DEVMODE=%#o", 0666);
    return 0;
}

static void __exit fli_pcie_exit(void)
{
   int   idx;

   pr_debug("Unloading (%s)\n", DEVICE_NAME);

   pci_unregister_driver(&pci_driver);
   unregister_chrdev_region(fli_pcie_dev_num, NUM_FLI_DEVICE_TOTAL);
   class_destroy(fli_class);

   fli_cam_exit();
#if 0
   fli_image_exit();
#endif

   for (idx=0; idx<NUM_FLI_PCIE_DEVICE; idx++)
   {
      kfree(pFli_dev_data[idx]);
   }

} // fli_pcie_exit


module_init(fli_pcie_init);
module_exit(fli_pcie_exit);
MODULE_LICENSE("GPL");
