//******************************************************************
/// @file flipcie_regblocks.h  
/// FLI PCIe Interface Board FPGA Register block Definitions for use
/// within the flipcie device driver
///
/// @copyright 2018 Finger Lakes Instrumentation LLC, All rights reserved
///
/// This file defines the PCIe memory mapped FPGA register block structs
/// used internally to the flipcie device driver.
///
//******************************************************************
#pragma once

#if defined(_WIN32) || defined(_WINDOWS)
#include "flipcie_types.h"
#else
#include <linux/types.h>
#endif

//******************************************************************
// Convenience Definitions
//******************************************************************

//******************************************************************
// FPGA Register block struct Definitions
//******************************************************************

//------------------------------------------------------------------
/**
* @name BAR 0 QSFP Interface Register Fields
*/
///@{ Start of QSFP Interface Register field definitions

// QSFP FPGA Registers
#pragma pack(1)
typedef struct
{
    uint32_t identifier;            ///< 0x00 Device ID 0x464C4902
    uint32_t version;               ///< 0x04 FPGA version 0x00MMmmbb
    uint32_t reg08;                 ///< 0x08 unused
    uint32_t reg0C;                 ///< 0x0C unused
    uint32_t ledControl;            ///< 0x10 LED control (deprecated)
    uint32_t reg14;                 ///< 0x14 unused
    uint32_t reg18;                 ///< 0x18 unused
    uint32_t reg1C;                 ///< 0x1C unused
    uint32_t qsfpCtrl;              ///< 0x20 QSFP control/status
    uint32_t trafficGenCtrl;        ///< 0x24 QSFP Traffic gen control/status
    uint32_t trafficGenNframes;     ///< 0x28 QSFP Traffic gen number of frames
    uint32_t trafficGenData;        ///< 0x2C QSFP Traffic gen initial data value
    uint32_t qsfpGoodPackets;       ///< 0x30 QSFP Good packet count (/1024)
    uint32_t qsfpBadPackets;        ///< 0x34 QSFP Bad packet count
    uint32_t trafficGenLenDly;      ///< 0x38 QSFP Traffic gen frame len (kB), delay (10uS)
    uint32_t reg3C;                 ///< 0x3C unused
    uint32_t dataCtrl;              ///< 0x40 host data interface control/status
    uint32_t dataXmitFifo;          ///< 0x44 host data interface transmit fifo
    uint32_t dataRecvFifo;          ///< 0x48 host data interface receive fifo
    uint32_t reg4C;                 ///< 0x4C unused
    uint32_t reserved[40];          ///< 0x50-0xEC unused
    uint32_t scratchPad0;           ///< 0xF0 Scratchpad register 0
    uint32_t scratchPad1;           ///< 0xF4 Scratchpad register 1
    uint32_t scratchPad2;           ///< 0xF8 Scratchpad register 2
    uint32_t scratchPad3;           ///< 0xFC Scratchpad register 3
} QSFP_REGS;
#pragma pack()

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 1 C2H Register Fields
*/
///@{ Start of C2H Register field definitions

// C2H XDMA Engine Registers
#pragma pack(1)
typedef struct 
{
    uint32_t identifier;            ///< 0x00 Engine Channel Identifier
    uint32_t control;               ///< 0x04 Engine Control Registger
    uint32_t controlW1S;            ///< 0x08 Ctrl Reg, write 1 to set
    uint32_t controlW1C;            ///< 0x0C Ctrl Reg, write 1 to clear
    uint32_t reserved1[12];         ///< 0x10-0x3C unused
    uint32_t status;                ///< 0x40 Engine Status Register
    uint32_t statusRC;              ///< 0x44 Sts Reg, clear on read
    uint32_t completedDescCount;    ///< 0x48 number of completed descriptors
    uint32_t alignments;            ///< 0x4C Memory alignment requirements
    uint32_t reserved2[12];         ///< 0x50-0x7C unused
    uint32_t reg80;                 ///< 0x80 unused
    uint32_t reg84;                 ///< 0x84 unused
    uint32_t pollModeWbLo;          ///< 0x88 Poll Mode write back address low word
    uint32_t pollModeWbHi;          ///< 0x8C Poll Mode write back address high word
    uint32_t intEnableMask;         ///< 0x90 Interrupt enable mask register
    uint32_t intEnableMaskW1S;      ///< 0x94 Int enable mask write 1 to set
    uint32_t intEnableMaskW1C;      ///< 0x98 Int enable mask write 1 to clear
    uint32_t reg9C;                 ///< 0x9C unused
    uint32_t reserved3[8];          ///< 0xA0-BC unused
    uint32_t perfCtrl;              ///< 0xc0 Performance Monitor control register
    uint32_t perfCycLo;             ///< 0xc4 Perf Monitor Cycle Count low word
    uint32_t perfCycHi;             ///< 0xc8 Perf monitor cycle count high word
    uint32_t perfDatLo;             ///< 0xcc Perf monitor data count low word
    uint32_t perfDatHi;             ///< 0xd0 Perf monitor data count high word
} XDMA_ENGINE_REGS;
#pragma pack()

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 2 User Interrupt Register Fields
*/
///@{ Start of Interrupt Register field definitions

// Interrupt Registers
#pragma pack(1)
typedef struct 
{
    uint32_t identifier;            ///< 0x00 Interrupt block identifier
    uint32_t userIntEnable;         ///< 0x04 User Interrupt Enable mask
    uint32_t userIntEnableW1S;      ///< 0x08 User Int Enable write 1 to set
    uint32_t userIntEnableW1C;      ///< 0x0C User Int Enable write 1 to clear
    uint32_t channelIntEnable;      ///< 0x10 Channel Interrupt Enable mask
    uint32_t channelIntEnableW1S;   ///< 0x14 Chan Int Enable write 1 to set
    uint32_t channelIntEnableW1C;   ///< 0x18 Chan Int Enable write 1 to clear
    uint32_t reg1C;                 ///< 0x1C unused
    uint32_t reserved1[8];          ///< 0x20-3c unused
    uint32_t userIntRequest;        ///< 0x40 User Interrupt Request
    uint32_t channelIntRequest;     ///< 0x44 Channel Interrupt Request
    uint32_t userIntPending;        ///< 0x48 User Interrupt Pending flags
    uint32_t channelIntPending;     ///< 0x4C Channel Interrupt Pending Flags
    uint32_t reserved2[12];         ///< 0x50-7c unused
    uint32_t userVector[4];         ///< 0x80-0x8c User Interrupt Vectors
    uint32_t reserved3[4];          ///< 90-9C unused
    uint32_t channelVector[2];      ///< 0xa0-0xA4 Channel Interrupt Vectors
} XDMA_IRQ_REGS;
#pragma pack()

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 3 Config Register Fields
*/
///@{ Start of Config Register field definitions

// Configuration Registers
#pragma pack(1)
typedef struct 
{
    uint32_t identifier;            ///< 0x00 Config block identifier
    uint32_t busDev;                ///< 0x04 Bus Device
    uint32_t maxPayload;            ///< 0x08 PCIe Max Payload size
    uint32_t maxReadReq;            ///< 0x0C PCIe Max Read Request size
    uint32_t systemID;              ///< 0x10 System ID
    uint32_t msiEnable;             ///< 0x14 MSI Enable
    uint32_t dataWidth;             ///< 0x18 PCIe Data width
    uint32_t pcieControl;           ///< 0x1C PCIe Control Register
    uint32_t reserved[8];           ///< 0x20-3C unused
    uint32_t userMaxPayload;        ///< 0x40 AXI User Max payload size
    uint32_t userMaxReadReq;        ///< 0x44 AXI User max Read Request size
    uint32_t writeFlushTimeout;     ///< 0x48 Write Flush timeout
} XDMA_CONFIG_REGS;
#pragma pack()

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 5 C2H SGDMA Register Fields
*/
///@{ Start of C2H SGDMA Register field definitions

// C2H Scatter-Gather DMA Registers
#pragma pack(1)
typedef struct 
{
    uint32_t identifier;            ///< 0x00 Block identifier
    uint32_t reg04;                 ///< 0x04 unused
    uint32_t reg08;                 ///< 0x08 unused
    uint32_t reg0C;                 ///< 0x0C unused
    uint32_t reserved[28];          ///< 0x10-0x7c
    uint32_t firstDescLo;           ///< 0x80 SGDMA Descriptor low address
    uint32_t firstDescHi;           ///< 0x84 SGDMA Descriptor high address
    uint32_t firstDescAdj;          ///< 0x88 SGDMA Descriptor adjacent count
    uint32_t DescCredits;           ///< 0x8C SGDMA Descriptor credit count
} XDMA_SGDMA_REGS;
#pragma pack()

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 6 Common SGDMA Register Fields
*/
///@{ Start of Common SGDMA Register field definitions
#pragma pack(1)
typedef struct
{
    uint32_t identifier;            ///< 0x00 Block identifier
    uint32_t reg04;                 ///< 0x04 unused
    uint32_t reg08;                 ///< 0x08 unused
    uint32_t reg0C;                 ///< 0x0C unused
    uint32_t descCtrl;              ///< 0x10 SGDMA Descriptor control
    uint32_t descCtrlW1S;           ///< 0x14 SGDMA Descriptor control, write 1 to set
    uint32_t descCtrlW1C;           ///< 0x18 SGDMA Descriptor control, write 1 to clear
    uint32_t reg1C;                 ///< 0x1C unused
    uint32_t descCreditEn;          ///< 0x20 SGDMA Descriptor credit mode enable
} XDMA_COM_REGS;
#pragma pack()

///@}


// Structure for polled mode descriptor writeback
// XDMA IP core writes number of completed descriptors to this memory, which the driver can then
// poll to detect transfer completion
typedef struct
{
    uint32_t uiCompletedDescCount;
    uint32_t uiReserved_1[7];
} XDMA_POLL_WB;

// DMA  Registers
typedef struct
{
    // register access
    volatile XDMA_ENGINE_REGS *pRegs;
    volatile XDMA_SGDMA_REGS  *pSgdma;
} XDMA_REGS;

//------------------------------------------------------------------
/**
* @name BAR 1 Block 8 MSI_X Interrupt Vector Register Fields
*/
///@{ Start of MSI_X Interrupt Vector Register field definitions
// MSI_X Vector table and Pending Bit Array
#pragma pack(1)
typedef struct
{
    uint32_t addressLo;             ///< vector message address low 32 bits
    uint32_t addressHi;             ///< vector message address high 32 bits
    uint32_t data;                  ///< vector message data
    uint32_t control;               ///< vector control
} MSI_X_VECTOR;

typedef MSI_X_VECTOR MSI_X_VECTOR_TABLE[32] ;
#pragma pack()

///@}

//******************************************************************
// End of file
//******************************************************************
