//******************************************************************
/// @file flipcie_regs.h  
/// FLI PCIe Interface Board FPGA Register Definitions.
///
/// @copyright 2017,2018 Finger Lakes Instrumentation LLC, All rights reserved
///
/// This file defines the PCIe memory mapped registers in the flipcie FPGA.
///
//******************************************************************
#pragma once

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_WIN32) || defined(_WINDOWS)
#ifdef _KERNEL_MODE
    #include "flipcie_types.h"
#else
    //#include <WinIoCtl.h>
    #include <stdint.h>
#endif
#endif

//******************************************************************
// FPGA Register Definitions
//******************************************************************
//------------------------------------------------------------------
/**
 * @name FPGA PCIe Register Addresses
 *
 * The FLI PCIe interface board contains a number of memory mapped control registers
 * accessible vie two PCIe Bus Address Register (BAR) regions.
 *
 * BAR 0 contains registers specific to the FLI PCIe FPGA
 * BAR 1 contains registers that control the Xilinx XDMA core within the FPGA.
 *
 * Each register is referenced by a 32-bit register address.
 * <pre>
 * - Address bits 31:17  : Reserved, must be 0
 * - Address bit     16  : BAR number
 * - Address bit  15:12  : Register block offset (always 0 for BAR 0)
 * - Address bits 11:08  : Reserved, must be 0
 * - Address bits 07:00  : Register byte address (bits 1:0 must be 0)
 * </pre>
 *
 * All registers are 32 bits wide, and support only 32-bit accesses.
 *
 * The following shorthand is used to define the access types supported by each register
 * <pre>
 * RO  : Read Only
 * WO  : Write Only
 * RW  : Read/Write
 * RC  : Clear on Read
 * W1C : Write 1 to clear a bit
 * W1S : Write 1 to set a bit
 * </pre>
 *
 */
 
///@{ BAR numbers and Register Block offsets
/// The REG_BAR0 and REG_BAR1 defines are NOT offsets.
/// They are a way to define register addresses with the bar number encoded
/// in them.  All register offsets are less than 16 bits.  We use bit 16
/// in the register definitions as the BAR identifier.  What this gets us is the
/// ability to have a single read register ioctl and a single write register ioctl.
/// The application in user land does not need to know in which physical BAR the
/// register resides.  It just uses the #define here for the register.
///
/// In the case of the Windows driver, this the driver knows this as well and
/// extracts the BAR number to ensure it uses the correct physical BAR for the access.
/// Similarly, it gets the actual register offset by simply masking off the BAR identifier.
///
/// For the Linux side of things, the user library does this masking and extracting
/// as the read and write ioctls there were implemented with the bar number included.  If the
/// Linux driver wishes to use these defines in its code, it must of course mask and extract
/// as appropriate.
#define REG_BAR0                (0x00000)
#define REG_BAR1                (0x10000)

#define REG_BAR_MASK            (0x10000)
#define REG_ADDR_MASK           (0x0FFFC)

#define GET_REG_BAR_NUM(addr) (((addr) & REG_BAR_MASK)>>16)
#define GET_REG_ADDR(addr) (((addr) & REG_ADDR_MASK))

#define OFFSET_QSFP             0x0000      ///< BAR offset for QSFP Registers
#define OFFSET_ISP              0x0000      ///< BAR offset for ISP related registers
#define OFFSET_C2H              0x1000      ///< BAR offset for C2H DMA dmaEngine registers
#define OFFSET_IRQ              0x2000      ///< BAR offset for Interrupt registers
#define OFFSET_CFG              0x3000      ///< BAR offset for PCIe Config registers
#define OFFSET_SGDMA            0x5000      ///< BAR offset for C2H SGDMA descriptor registers
#define OFFSET_COM              0x6000      ///< BAR offset for common SGDMA descriptor control
#define OFFSET_MSIX             0x8000      ///< BAR offset for MSI-X Interrupt vector registers

#define REG_QSFP                (REG_BAR0 + OFFSET_QSFP)    ///< QSFP interface specific registers
#define REG_C2H                 (REG_BAR1 + OFFSET_C2H)     ///< Card to Host DMA dmaEngine registers
#define REG_IRQ                 (REG_BAR1 + OFFSET_IRQ)     ///< Interrupt registers
#define REG_CFG                 (REG_BAR1 + OFFSET_CFG)     ///< DMA/PCie Config registers
#define REG_SGDMA               (REG_BAR1 + OFFSET_SGDMA)   ///< C2H SGDMA descriptor registers
#define REG_COM                 (REG_BAR1 + OFFSET_COM)     ///< Common SGDMA descriptor registers
#define REG_MSIX                (REG_BAR1 + OFFSET_MSIX)    ///< MSI-X Interrupt vector registers

///@}

///@{ BAR 0 QSFP Specific Register addresses
#define REG_ID                  (REG_QSFP + 0x0000) ///< (RO) FPGA ID Register (0x464C4902)
#define REG_VERSION             (REG_QSFP + 0x0004) ///< (RO) FPGA Version Register
#define REG_FIBRESIDE_VERSION   (REG_VERSION)       ///< (RO) FPGA Version Register
#define REG_PCIESIDE_VERSION    (0x1004)            ///< (RO) FPGA Version Register
#define REG_LED_CTRL            (REG_QSFP + 0x0010) ///< (RW) LED Control (deprecated)

#define REG_QSFP_CTRL           (REG_QSFP + 0x0020) ///< (RW) QSFP Fiber Optic interface Control
//#define REG_TR_GEN_CTRL         (REG_QSFP + 0x0024) ///< (RW) QSFP Traffic Generator Control
//#define REG_TR_GEN_NFRAMES      (REG_QSFP + 0x0028) ///< (RW) QSFP Traffic Generator number of frames to generate
//#define REG_TR_GEN_DATA         (REG_QSFP + 0x002C) ///< (RW) QSFP Traffic Generator initial data value
#define REG_QSFP_GOOD_PKT_COUNT (REG_QSFP + 0x0030) ///< (RW) QSFP Good Packet Count (in units of 1024 bytes)
#define REG_QSFP_BAD_PKT_COUNT  (REG_QSFP + 0x0034) ///< (RW) QSFP Bad Packet Count (in units of 1024 bytes)
//#define REG_TR_GEN_LEN_DLY      (REG_QSFP + 0x0038) ///< (RW) QSFP Traffic Generator Frame length and delay


#define REG_DATA_CTRL                     (REG_QSFP + 0x0040) ///< (RW) Host Data Interface Control/Status
#define REG_DATA_XMIT_FIFO                (REG_QSFP + 0x0044) ///< (WO) Host Data Interface Transmit Fifo
#define REG_DATA_RCV_FIFO                 (REG_QSFP + 0x0048) ///< (RO) Host Data Interface Receive Fifo
#define REG_DATA_V3CONNECTION_CTRL        (REG_QSFP + 0x0070) ///< (RW) V3 Card connection Status and Control
#define REG_DATA_IMAGE_CTRL               (REG_QSFP + 0x0060) ///< (RW) Image Control
#define REG_WATER_MARKS                   (REG_QSFP + 0x00A0) ///< Fibre transfer water marks

#define IMAGE_CTRL_LINE_BYTE_LENGTH_MASK  (0x0000FFFF)
#define IMAGE_CTRL_META_DTA_LENGTH_MASK   (0x0FFF0000)
#define IMAGE_CTRL_IMAGE_CONFIG_MASK      (0xF0000000)   // 0= 1 image per picture, 1 = 2 image per pic (essentially LDR vs HDR)
#define IMAGE_CTRL_LINE_BYTE_LENGTH_SHIFT (0)
#define IMAGE_CTRL_META_DTA_LENGTH_SHIFT  (16)
#define IMAGE_CTRL_IMAGE_CONFIG_SHIFT     (28)           // 0= 1 image per picture, 1 = 2 image per pic (essentially LDR vs HDR)
#define IMAGE_CTRL_IMAGE_HDR_BIT      (BIT_N(IMAGE_CTRL_IMAGE_CONFIG_SHIFT))

///@{ BAR 0 ISP registers
#define REG_ISP_DEV_OFFSET_DELTA (0x1000)    /// Offset for Fibre device side ISP registers- see m_uiFPGADeviceOffset
#define REG_ISP_FIBRE_DEV_OFFSET (0 * REG_ISP_DEV_OFFSET_DELTA)    /// Offset for Fibre device side ISP registers
#define REG_ISP_PCIE_DEV_OFFSET  (1 * REG_ISP_DEV_OFFSET_DELTA)    /// Offset for PCIE device side ISP registers
#define REG_ISP_CONTROL          (REG_QSFP + 0x00E0)    ///< (RW)  Control Register
#define REG_ISP_WRITE_FIFO       (REG_QSFP + 0x00E4)    ///< (WO)  Write Fifo
#define REG_ISP_READ_FIFO        (REG_QSFP + 0x00E8)    ///< (RO)  Read Fifo
#define REG_ISP_STATUS           (REG_QSFP + 0x00EC)    ///< (RW)  Control Register
#define REG_ISP_READ_FIFO_COUNT  (REG_QSFP + 0x00D0)    ///< (RW)  Control Register
#define REG_ISP_WRITE_FIFO_COUNT (REG_QSFP + 0x00D4)    ///< (RW)  Control Register

#define REG_PCIE_TRGEN_CTRL       (REG_ISP_DEV_OFFSET_DELTA + 0x20)
#define REG_PCIE_TRGEN_NUMFRAMES  (REG_ISP_DEV_OFFSET_DELTA + 0x24)
#define REG_PCIE_TRGEN_INITVALUE  (REG_ISP_DEV_OFFSET_DELTA + 0x28)
#define REG_PCIE_TRGEN_FRAMELEN   (REG_ISP_DEV_OFFSET_DELTA + 0x2C)
#define REG_FIBRE_TRGEN_CTRL      (0xB0)
#define REG_FIBRE_TRGEN_NUMFRAMES (0xB4)
#define REG_FIBRE_TRGEN_INITVALUE (0xB8)
#define REG_FIBRE_TRGEN_FRAMELEN  (0xBC)

#define REG_PCIE_PG_CTRL             (0x1020)
#define REG_PCIE_PG_NUM_FRAMES       (0x1024)
#define REG_PCIE_PG_START_VAL        (0x1028)
#define REG_PCIE_PG_FRAME_SIZE       (0x102C)
#define REG_PCIE_REFFRAME_START_ADDR (0x1040)
#define REG_PCIE_REFFRAME_LENGTH     (0x1044)
#define REG_PCIE_REFFRAME_ENABLE     (0x1048)
#define REG_PCIE_REFFRAME_STATUS     (0x104C)
#define REG_PCIE_REFFRAME_DATA       (0x1050)
#define REG_PCIE_MERGE_THRESHOLDS    (0x1070) ///< Merging thresholds
#define REG_PCIE_REFFRAME_STATUS_BUSY       (0x0001)
#define REG_PCIE_REFFRAME_STATUS_LOADED     (0x0002)
#define REG_PCIE_REFFRAME_STATUS_OVERFLOW   (0x0004)
#define REG_PCIE_REFFRAME_STATUS_TREADY     (0x0008)
#define REG_PCIE_REFFRAME_STATUS_COUNT_MASK (0xFFFF0000)

// Other Version 2 registers
#define REG_FIBRE_BUFFER_BYPASS   (0xA4)
#define REG_FIBRE_FRAME_SIZE      (0xA8)

#define REG_SCRATCHPAD_0         (REG_QSFP + 0x00F0) ///< (RW) General purpose register 0
#define REG_SCRATCHPAD_1         (REG_QSFP + 0x00F4) ///< (RW) General purpose register 1
#define REG_SCRATCHPAD_2         (REG_QSFP + 0x00F8) ///< (RW) General purpose register 2
#define REG_SCRATCHPAD_3         (REG_QSFP + 0x00FC) ///< (RW) General purpose register 3

///@}

///@{ BAR 1 Block 1 C2H Register addresses
#define REG_C2H_ID              (REG_C2H + 0x00)    ///< (RO)  C2H Channel register block Identifier

#define REG_C2H_CTRL            (REG_C2H + 0x04)    ///< (RW)  C2H Channel Control
#define REG_C2H_CTRL_1S         (REG_C2H + 0x08)    ///< (W1S) C2H Channel Control, 1 to set
#define REG_C2H_CTRL_1C         (REG_C2H + 0x0C)    ///< (W1C) C2H Channel Control, 1 to clear

#define REG_C2H_STS_1C          (REG_C2H + 0x40)    ///< (W1C) C2H Channel Status 
#define REG_C2H_STS_RC          (REG_C2H + 0x44)    ///< (RC)  C2H Channel Status 

#define REG_C2H_COMP_COUNT      (REG_C2H + 0x48)    ///< (RO)  C2H Completed Descriptor Count
#define REG_C2H_ALIGN           (REG_C2H + 0x4C)    ///< (RO)  C2H address and length alignment requirements
#define REG_C2H_WB_ADDR_LO      (REG_C2H + 0x88)    ///< (RW)  C2H Poll mode writeback address low 32 bits
#define REG_C2H_WB_ADDR_HI      (REG_C2H + 0x8C)    ///< (RW)  C2H Poll mode writeback address high 32 bits

#define REG_C2H_INT_MASK        (REG_C2H + 0x90)    ///< (RW)  C2H Interrupt enable mask
#define REG_C2H_INT_MASK_1S     (REG_C2H + 0x94)    ///< (W1S) C2H Interrupt enable mask, 1 to set
#define REG_C2H_INT_MASK_1C     (REG_C2H + 0x98)    ///< (W1C) C2H Interrupt enable mask, 1 to clear

#define REG_C2H_PERF_CTRL       (REG_C2H + 0xC0)    ///< (RW)  C2H Performance monitor control
#define REG_C2H_PERF_CYCLES_LO  (REG_C2H + 0xC4)    ///< (RO)  C2H Performance monitor clock cycle count low 32 bits
#define REG_C2H_PERF_CYCLES_HI  (REG_C2H + 0xC8)    ///< (RO)  C2H Performance monitor clock cycle count high10 bits
#define REG_C2H_PERF_DATA_LO    (REG_C2H + 0xCC)    ///< (RO)  C2H Performance monitor data count low 32 bits
#define REG_C2H_PERF_DATA_HI    (REG_C2H + 0xD0)    ///< (RO)  C2H Performance monitor data count high 10 bits
///@}

///@{ BAR 1 Block 2 User IRQ Register addresses
#define REG_IRQ_ID              (REG_IRQ + 0x00)    ///< (RO)  IRQ Register block Identifier

#define REG_IRQ_USR_MASK        (REG_IRQ + 0x04)    ///< (RW)  IRQ User interrupt enable mask
#define REG_IRQ_USR_MASK_1S     (REG_IRQ + 0x08)    ///< (W1S) IRQ User interrupt enable mask, 1 to set
#define REG_IRQ_USR_MASK_1C     (REG_IRQ + 0x0C)    ///< (W1C) IRQ User interrupt enable mask, 1 to clear

#define REG_IRQ_CHAN_MASK       (REG_IRQ + 0x10)    ///< (RW)  IRQ Channel interrupt enable mask
#define REG_IRQ_CHAN_MASK_1S    (REG_IRQ + 0x14)    ///< (W1S) IRQ Channel interrupt enable mask, 1 to set
#define REG_IRQ_CHAN_MASK_1C    (REG_IRQ + 0x18)    ///< (W1C) IRQ Channel interrupt enable mask, 1 to clear

#define REG_IRQ_USR_REQ         (REG_IRQ + 0x40)    ///< (RO)  IRQ User interrupt request
#define REG_IRQ_CHAN_REQ        (REG_IRQ + 0x44)    ///< (RO)  IRQ Channel interrupt request

#define REG_IRQ_USR_PEND        (REG_IRQ + 0x48)    ///< (RO)  IRQ User interrupt pending
#define REG_IRQ_CHAN_PEND       (REG_IRQ + 0x4C)    ///< (RO)  IRQ Channel interrupt pending

#define REG_IRQ_USR_VECTOR_0_3  (REG_IRQ + 0x80)    ///< (RW)  IRQ User interrupt vector 3
#define REG_IRQ_USR_VECTOR_4_7  (REG_IRQ + 0x84)    ///< (RW)  IRQ User interrupt vector 2
#define REG_IRQ_USR_VECTOR_8_B  (REG_IRQ + 0x88)    ///< (RW)  IRQ User interrupt vector 1
#define REG_IRQ_USR_VECTOR_C_F  (REG_IRQ + 0x8C)    ///< (RW)  IRQ User interrupt vector 0

#define REG_IRQ_CHAN_VECTOR_0_3 (REG_IRQ + 0xA0)    ///< (RW)  IRQ Channel interrupt vectors 4 thru 7
#define REG_IRQ_CHAN_VECTOR_4_7 (REG_IRQ + 0xA4)    ///< (RW)  IRQ Channel interrupt vectors 8 thru 11
///@}

///@{ BAR 1 Block 3 Config Register addresses
#define REG_CFG_ID                  (REG_CFG + 0x00)    ///< (RO)  CFG Register block identifier
#define REG_CFG_BUSDEV              (REG_CFG + 0x04)    ///< (RO)  CFG PCIe Bus, Device, and Function
#define REG_CFG_PCIE_MAX_PAYLOAD    (REG_CFG + 0x08)    ///< (RO)  CFG PCIe max payload size
#define REG_CFG_PCIE_MAX_RD_REQ     (REG_CFG + 0x0C)    ///< (RO)  CFG PCIe max read request size
#define REG_CFG_SYS_ID              (REG_CFG + 0x10)    ///< (RO)  CFG System ID
#define REG_CFG_MSI_ENABLE          (REG_CFG + 0x14)    ///< (RO)  CFG MSI enable
#define REG_CFG_DATA_WIDTH          (REG_CFG + 0x18)    ///< (RO)  CFG PCIe data width
#define REG_CFG_PCIE_CTRL           (REG_CFG + 0x1C)    ///< (Rw)  CFG PCIe Control
#define REG_CFG_AXI_MAX_PAYLOAD     (REG_CFG + 0x40)    ///< (RW)  CFG AXI User max payload size
#define REG_CFG_AXI_MAX_RD_REQ      (REG_CFG + 0x44)    ///< (RW)  CFG AXI User max read request size
#define REG_CFG_WRT_FLUSH_TIMEOUT   (REG_CFG + 0x60)    ///< (RW)  CFG Write Flush timeout
///@}

///@{ BAR 1 Block 5 C2H SGDMA Register addresses
#define REG_SGD_ID                  (REG_SGDMA + 0x00)  ///< (RO)  DMA Register block identifier
#define REG_SGD_DESC_LOW            (REG_SGDMA + 0x80)  ///< (RO)  DMA Descriptor table address, low 32 bits
#define REG_SGD_DESC_HIGH           (REG_SGDMA + 0x84)  ///< (RO)  DMA Descriptor table address, high 32 bits
#define REG_SGD_DESC_ADJACENT       (REG_SGDMA + 0x88)  ///< (RO)  DMA number of adjacent descriptors following the start
#define REG_SGD_DESC_CREDITS        (REG_SGDMA + 0x8C)  ///< (RW)  DMA Descriptor credits -- ??
///@}

///@{ BAR 1 Block 6 Common SGDMA Register addresses
#define REG_COM_ID                  (REG_COM + 0x00)    ///< (RO)  Common register block identifier
#define REG_COM_CTRL                (REG_COM + 0x10)    ///< (RW)  Descriptor control
#define REG_COM_CTRL_1S             (REG_COM + 0x14)    ///< (W1S) Descriptor control, 1 to set
#define REG_COM_CTRL_1C             (REG_COM + 0x18)    ///< (W1C) Descriptor control, 1 to clear
#define REG_COM_DESC_MODE           (REG_COM + 0x20)    ///< (RW)  Descriptor credit mode
#define REG_COM_DESC_MODE_1S        (REG_COM + 0x24)    ///< (W1S) Descriptor credit mode, 1 to set 
#define REG_COM_DESC_MODE_1C        (REG_COM + 0x28)    ///< (W1C) Descriptor credit mode, 1 to clear 
///@}

///@{ BAR 1 Block 8 MSI-X Vector Table registers
#define REG_MSI_VECT_00_LO          (REG_SGDMA + 0x000) ///< (RW)  MSI_X Vector 0 Address, low 32 bits
#define REG_MSI_VECT_00_HI          (REG_SGDMA + 0x004) ///< (RW)  MSI_X Vector 0 Address, high 32 bits
#define REG_MSI_VECT_00_DATA        (REG_SGDMA + 0x008) ///< (RW)  MSI_X Vector 0 Message data
#define REG_MSI_VECT_00_CTRL        (REG_SGDMA + 0x00C) ///< (RW)  MSI_X Vector 0 Control

#define REG_MSI_VECT_01_LO          (REG_SGDMA + 0x010) ///< (RW)  MSI_X Vector 1 Address, low 32 bits
#define REG_MSI_VECT_01_HI          (REG_SGDMA + 0x014) ///< (RW)  MSI_X Vector 1 Address, high 32 bits
#define REG_MSI_VECT_01_DATA        (REG_SGDMA + 0x018) ///< (RW)  MSI_X Vector 1 Message data
#define REG_MSI_VECT_01_CTRL        (REG_SGDMA + 0x01C) ///< (RW)  MSI_X Vector 1 Control
// ... add other vectors here if needed
#define REG_MSI_VECT_1F_LO          (REG_SGDMA + 0x1F0) ///< (RW)  MSI_X Vector 31 Address, low 32 bits
#define REG_MSI_VECT_1F_HI          (REG_SGDMA + 0x1F4) ///< (RW)  MSI_X Vector 31 Address, high 32 bits
#define REG_MSI_VECT_1F_DATA        (REG_SGDMA + 0x1F8) ///< (RW)  MSI_X Vector 31 Message data
#define REG_MSI_VECT_1F_CTRL        (REG_SGDMA + 0x1FC) ///< (RW)  MSI_X Vector 31 Control

///@}

#define REG_TRGEN_CONFIG_LENGTH_1             (0x60)
#define REG_TRGEN_CONFIG_LENGTH_2             (0x64)
#define REG_TRGEN_CONFIG_LOW_GAIN             (0x68)
#define REG_TRGEN_CONFIG_HIGH_GAIN            (0x6C)
#define REG_TRGEN_CONFIG_ACTUAL_LENGTH_BYTES  (0x70)
#define REG_FPGA_DDR_CONTROL                  (0xA4)
#define REG_FPGA_IMAGE_PROCESSIING            (0x1038)
#define REG_FPGA_FRAME_ROW_OFFSET_START       (0x1074)
#define REG_FPGA_FIBRE_TEMPERATURE            (0xC0)
#define REG_FPGA_PCIE_TEMPERATURE             (0x10C0)

//------------------------------------------------------------------
/**
* @name Device Driver Pseudo Register Addresses
*/
///@{ Start of Device Driver Pseudo Register addresses
#define REG_DRV_ID                  0x0000              ///< (RO) Device Driver Identifier
#define REG_DRV_VER                 0x0004              ///< (RO) Device Driver Version MMmmPPBB
#define REG_DRV_STS                 0x0008              ///< (RW) Device Driver Status
#define REG_DRV_CTRL                0x000C              ///< (RW) Device Driver Control
#define REG_DRV_MAX                 0x0010              ///< number of bytes available

///@}

//------------------------------------------------------------------
/**
* @name BAR 0 QSFP Interface Register Fields
*/
///@{ Start of QSFP Interface Register field definitions

// QSFP Control Register fields
#define QSFP_CR_DATA_RX_OVERFLOW    BIT_N(31)
#define QSFP_CR_DATA_TX_OVERFLOW    BIT_N(30)
#define QSFP_CR_IMAGE_RX_OVERFLOW   BIT_N(29)
#define QSFP_CR_IMAGE_TX_OVERFLOW   BIT_N(28)
#define QSFP_CR_XCVR_FAULT          BIT_N(27)
#define QSFP_CR_HARD_ERROR          BIT_N(26)
#define QSFP_CR_SOFT_ERROR          BIT_N(25)
#define QSFP_CR_FRAME_ERROR         BIT_N(24)
#define QSFP_CR_LANE3_UP            BIT_N(23)
#define QSFP_CR_LANE2_UP            BIT_N(22)
#define QSFP_CR_LANE1_UP            BIT_N(21)
#define QSFP_CR_LANE0_UP            BIT_N(20)
#define QSFP_CR_CHANNEL_UP          BIT_N(19)
#define QSFP_CR_REF_CLOCK_LOCK      BIT_N(18)
#define QSFP_CR_TX_RESET_DONE       BIT_N(17)
#define QSFP_CR_RX_RESET_DONE       BIT_N(16)
#define QSFP_CR_XON_STATE_SENT      BIT_N(15)
#define QSFP_CR_XON_STATE_RCVD      BIT_N(14)
#define QSFP_CR_XCVR_PRESENT        BIT_N(13)
#define QSFP_CR_INJECT_ERROR        BIT_N( 7)
#define QSFP_CR_DATA_FIFO_RESET     BIT_N( 6)
#define QSFP_CR_IMAGE_FIFO_RESET    BIT_N( 5)
#define QSFP_CR_SOFT_RESET          BIT_N( 4)
#define QSFP_CR_POWER_DOWN          BIT_N( 3)
#define QSFP_CR_LOOPBACK_2          BIT_N( 2)  ///< 0:Near end, 1: Far end
#define QSFP_CR_LOOPBACK_1          BIT_N( 1)  ///< 0:Physical Coding Sublayer, 1:Physical Media Attachment
#define QSFP_CR_LOOPBACK_0          BIT_N( 0)

// QSFP Reset is complete
#define QSFP_CR_RESET_DONE         (QSFP_CR_RX_RESET_DONE  | \
                                    QSFP_CR_TX_RESET_DONE  | \
                                    QSFP_CR_REF_CLOCK_LOCK) 

// All QSFP lanes are operational
#define QSFP_CR_ALL_LANES_UP       (QSFP_CR_LANE0_UP | \
                                    QSFP_CR_LANE1_UP | \
                                    QSFP_CR_LANE2_UP | \
                                    QSFP_CR_LANE3_UP )
#define QSFP_CR_LANES_AND_CHANNEL_UP       (QSFP_CR_LANE0_UP | \
                                            QSFP_CR_LANE1_UP | \
                                            QSFP_CR_LANE2_UP | \
                                            QSFP_CR_LANE3_UP | \
                                            QSFP_CR_CHANNEL_UP)
// V3 card Status and control bits
#define QSFP_V3CR_CHANNEL_SELECT_BIT      BIT_N(15)

#define QSFO_V3CR_TOP_CHANNEL_MASK        (0x00007F00)
#define QSFP_V3CR_TOP_CHANNEL_INTN        BIT_N(14)
#define QSFP_V3CR_TOP_CHANNEL_UP          BIT_N(13)
#define QSFP_V3CR_TOP_LANE3_UP            BIT_N(12)
#define QSFP_V3CR_TOP_LANE2_UP            BIT_N(11)
#define QSFP_V3CR_TOP_LANE1_UP            BIT_N(10)
#define QSFP_V3CR_TOP_LANE0_UP            BIT_N(9)
#define QSFP_V3CR_TOP_QSFP_MODULE_PRESENT BIT_N(8)
#define QSFP_V3CR_TOP_LANES_AND_CHANNEL_UP (QSFP_V3CR_TOP_LANE0_UP | \
                                             QSFP_V3CR_TOP_LANE1_UP | \
                                             QSFP_V3CR_TOP_LANE2_UP | \
                                             QSFP_V3CR_TOP_LANE3_UP | \
                                             QSFP_V3CR_TOP_CHANNEL_UP)

#define QSFO_V3CR_BOTTOM_CHANNEL_MASK        (0x0000007F)
#define QSFP_V3CR_BOTTOM_CHANNEL_INTN         BIT_N(6)
#define QSFP_V3CR_BOTTOM_CHANNEL_UP           BIT_N(5)
#define QSFP_V3CR_BOTTOM_LANE3_UP             BIT_N(4)
#define QSFP_V3CR_BOTTOM_LANE2_UP             BIT_N(3)
#define QSFP_V3CR_BOTTOM_LANE1_UP             BIT_N(2)
#define QSFP_V3CR_BOTTOM_LANE0_UP             BIT_N(1)
#define QSFP_V3CR_BOTTOM_QSFP_MODULE_PRESENT  BIT_N(0)
#define QSFP_V3CR_BOTTOM_LANES_AND_CHANNEL_UP (QSFP_V3CR_BOTTOM_LANE0_UP | \
                                              QSFP_V3CR_BOTTOM_LANE1_UP | \
                                              QSFP_V3CR_BOTTOM_LANE2_UP | \
                                              QSFP_V3CR_BOTTOM_LANE3_UP | \
                                              QSFP_V3CR_BOTTOM_CHANNEL_UP)

// QSFP interface should be operational
#define QSFP_CR_READY              (QSFP_CR_RESET_DONE | \
                                    QSFP_CR_ALL_LANES_UP)

#define QSFP_CR_ERRORS             (QSFP_CR_FRAME_ERROR | \
                                    QSFP_CR_SOFT_ERROR  | \
                                    QSFP_CR_HARD_ERROR  | \
                                    QSFP_CR_XCVR_FAULT )

// QSFP Traffic Generator Register fields
#define QSFP_TG_CHECKER_ERROR       BIT_N(19)
#define QSFP_TG_DONE                BIT_N(18)
#define QSFP_TG_RX_RUNNING          BIT_N(17)
#define QSFP_TG_TX_RUNNING          BIT_N(16)
#define QSFP_TG_RX_MUX              BIT_N( 6)
#define QSFP_TG_IMAGE_TAG           BIT_N( 5)
#define QSFP_TG_GAP_ENABLE          BIT_N( 4)   ///< FPGA V1.6.1 and earlier only
#define QSFP_TG_CONTINUOUS          BIT_N( 3)   ///< FPGA V1.6.1 and earlier only
#define QSFP_TG_TX_ENABLE           BIT_N( 2)
#define QSFP_TG_METADATA_ENABLE     BIT_N( 2)   ///< V2 Fibre side only
#define QSFP_TG_RX_ENABLE           BIT_N( 1)
#define QSFP_TG_ENABLE              BIT_N( 0)   ///< Enable traffic generator (rising edge)

// QSFP Data Interface Control Register fields
#define QSFP_DC_TX_MSG_PENDING      BIT_N(10)   ///< 1: message not yet picked up on other end
#define QSFP_DC_TX_FIFO_EMPTY       BIT_N( 9)   ///< 1: TX fifo is empty
#define QSFP_DC_RX_MSG_RCVD         BIT_N( 8)   ///< 1: message received from other end
#define QSFP_DC_TX_MSG_SEND         BIT_N( 0)   ///< Send message (rising edge)

#define QSFP_DC_RX_WORD_COUNT_OFFSET   (16)
#define QSFP_DC_RX_WORD_COUNT_MASK  (0x0FFF << QSFP_DC_RX_WORD_COUNT_OFFSET)
#define QSFP_DC_RX_WORD_COUNT(reg) \
    (((reg) & QSFP_DC_RX_WORD_COUNT_MASK) >> QSFP_DC_RX_WORD_COUNT_OFFSET)

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 1 C2H Register Fields
*/
///@{ Start of C2H Register field definitions

// Block identifier register fields
#define XDMA_ID_MASK                (0xfffUL<<20) ///< Location of XDMA identifier in register 0 
#define XDMA_ID                     (0x1fcUL<<20) ///< XDMA Identifier value
#define XDMA_ID_STREAM              BIT_N(15)    ///< 0=Memory mapped, 1=Streaming DMA

// C2H XDMA dmaEngine control register fields
#define XDMA_CR_RUN_BIT             BIT_N(0)    ///< 1 to start dmaEngine, 0 to stop
#define XDMA_CR_IE_DESC_STOPPED     BIT_N(1)    ///< 1 to log Status_Descriptor_Stopped
#define XDMA_CR_IE_DESC_COMPLETED   BIT_N(2)    ///< 1 to log Status_Descriptor_completed
#define XDMA_CR_IE_ALIGN_MISMATCH   BIT_N(3)    ///< 1 to log Status_Align_mismatch 
#define XDMA_CR_IE_MAGIC_STOPPED    BIT_N(4)    ///< 1 to log Status_Magix_stopped
#define XDMA_CR_IE_INVALID_LENGTH   BIT_N(5)    ///< 1 to log Status_Invalid_length
#define XDMA_CR_IE_IDLE_STOPPED     BIT_N(6)    ///< 1 to log Status_Idle_stopped
#define XDMA_CR_IE_READ_ERROR       (0x1f<<9)   ///< 0x1f to log Status_read_error and stop if error
#define XDMA_CR_IE_WRITE_ERROR      (0x1f<<14)  ///< 0x1f to log Status_write_error and stop if error
#define XDMA_CR_IE_DESCRIPTOR_ERROR (0x1f<<19)  ///< 0x1f to log Status_Desc_error and stop if error
#define XDMA_CR_NON_INCR_ADDR       BIT_N(25)   ///< 1 for non-incrementing address mode (mm only)
#define XDMA_CR_POLL_MODE           BIT_N(26)   ///< 1 for to enable poll mode writeback
#define XDMA_CR_NO_METADATA         BIT_N(27)   ///< 1 to disable metadata writeback (stream only)

#define XDMA_CR_IE_ALL ( XDMA_CR_IE_DESC_STOPPED | XDMA_CR_IE_DESC_COMPLETED \
                       | XDMA_CR_IE_ALIGN_MISMATCH | XDMA_CR_IE_MAGIC_STOPPED \
                       | XDMA_CR_IE_INVALID_LENGTH | XDMA_CR_IE_READ_ERROR | XDMA_CR_IE_WRITE_ERROR\
                       | XDMA_CR_IE_DESCRIPTOR_ERROR)

// C2H XDMA dmaEngine status register fields
#define XDMA_SR_BUSY                BIT_N(0)    ///< 1 if DMA dmaEngine is busy
#define XDMA_SR_DESC_STOPPED        BIT_N(1)    ///< 1 if descriptor completed with STOP bit set
#define XDMA_SR_DESC_COMPLETED      BIT_N(2)    ///< 1 if descriptor completed with COMPLETE bit set
#define XDMA_SR_ALIGN_MISMATCH      BIT_N(3)    ///< 1 if src and dest are not aligned to each other
#define XDMA_SR_MAGIC_STOPPED       BIT_N(4)    ///< 1 if stopped due to descriptor with invalid magic 
#define XDMA_SR_FETCH_STOPPED       BIT_N(5)    ///< 1 if stopped due to descriptor with invalid length
#define XDMA_SR_IDLE_STOPPED        BIT_N(6)    ///< 1 if dmaEngine is idle
#define XDMA_SR_READ_ERROR          (0x1f<<9)   ///< Read error flags
#define XDMA_SR_DESC_ERROR          (0x1f<<19)  ///< Descriptor error flags

#define XDMA_SR_ENGINE_STOPPED_OK   (0UL)
#define XDMA_SR_EXPECTED_ZERO       (XDMA_SR_BUSY | XDMA_SR_MAGIC_STOPPED | \
                                     XDMA_SR_FETCH_STOPPED | XDMA_SR_ALIGN_MISMATCH | \
                                     XDMA_SR_READ_ERROR | XDMA_SR_DESC_ERROR)
                                         
// C2H XDMA dmaEngine performance monitor control register fields
#define XDMA_PC_RUN                 BIT_N(2)   ///< 1:perf counters start when DMA CtrlRUN bit set
#define XDMA_PC_CLEAR               BIT_N(1)   ///< 1:Clear perf counters
#define XDMA_PC_AUTO                BIT_N(0)   ///< 1:Stop perf counters when descriptors are completed

// Poll Write Back  info
#define XDMA_WB_COUNT_MASK              (0x00ffffffUL)
#define XDMA_WB_ERR_MASK                (BIT_N(31))

///@}
#define FIBRE_BUFFER_BYPASS_ENABLE_BIT   (BIT_N(1))

//------------------------------------------------------------------
/**
* @name BAR 1 Block 2 User Interrupt Register Fields
*/
///@{ Start of Interrupt Register field definitions

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 3 Config Register Fields
*/
///@{ Start of Config Register field definitions

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 5 C2H SGDMA Register Fields
*/
///@{ Start of C2H SGDMA Register field definitions

///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 6 Common SGDMA Register Fields
*/
///@}

//------------------------------------------------------------------
/**
* @name BAR 1 Block 8 MSI_X Interrupt Vector Register Fields
*/
///@}

//******************************************************************
// End of file
//******************************************************************
#ifdef __cplusplus
} // end extern "C"
#endif //__cplusplus
